/* FILE: cmpIFCompiler.h
 *
 * PROGRAM: User Friendly Metadata
 *
 * DESCRIPTION: Header file for the UFM's Instructions File Compiler module.
 *
 * RCS INFO:
 *
 *	$Id: cmpIFCompiler.h,v 1.1 2002/05/08 15:20:20 yin Exp $
 *
 *	$Log: cmpIFCompiler.h,v $
 *	Revision 1.1  2002/05/08 15:20:20  yin
 *	Initial revision
 *
 *	Revision 1.4  2000/04/07 00:25:42  ToolsDev
 *	Made changes to display the ODL file header informations:
 *	added a new pointer *pHeader in the REP Structure.
 *	(Amer Rahabi)
 *
 * Revision 1.3  1999/04/06  15:54:11  stroup
 * Promotion of branch 1.2.0 (cleared wlint warnings) for release.
 *
 * Revision 1.2.0.1  1999/03/10  14:40:48  stroup
 * Correction of errors found by lint: changed datatype of some variables
 * from 'int' to 'unsigned int'.
 *
 * Revision 1.2  1999/02/03  18:34:26  jarva
 * Promotions of the 1.1.0 branch
 *
 * Revision 1.1.0.1  1999/01/26  17:13:59  jarva
 * Eliminated prototype for cmpComponentName.
 *
 * Revision 1.1  1999/01/06  20:40:42  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
#ifndef _cmpIFCompiler_header
#define _cmpIFCompiler_header

/* INCLUDED FILES: the following included files are here because of
 *     dependencies that exist within this header file.
*/
#include "sltLUT.h"
#include "odldef.h"
#include "ODL_Node.h"


/* ---------------------------------------------------------------------*/

/* EXTERN VARIABLES: the following variables may be used in other files.
 *    They are defined in cmpIFCompiler.c.
*/

extern sltLUT pMacroLUT;
extern sltLUT pAggByNameLUT;
extern sltLUT pAggByClassLUT;	
extern sltLUT pParByNameLUT;
extern sltLUT pParByClassLUT;
extern sltLUT pIDTranslationLUT;

/* ---------------------------------------------------------------------*/

/* ENUMERATED TYPES ____________________________________________________*/

/* The kinds of instructions found in the Instructions File */
enum InstructionKind {
	AGGREGATE_TRANSLATION, PARAMETER_TRANSLATION, IDENTIFIER_TRANSLATIONS,
		MACRO_DEFINITIONS, CLASS_DEFINITION, UNKNOWN_INSTRUCTION
};
typedef enum InstructionKind INST_KIND;

/* The various kinds of reportNodes.
 *     cmpSUBLIST means a sub list of reportNODEs, cmpDATA means that
 *     the node represents a "name = value_list" kind of thing,
 *     cmpTABLE means that the reportNode represents a table, and
 *     cmpIGNORE means that the node represents a component that is to
 *     be ignored.
*/
enum reportNode_kind { 
	cmpSUBLIST, cmpDATA, cmpTABLE, cmpIGNORE
};
typedef enum reportNode_kind RN_KIND;

/* STRUCTURE DEFINITIONS ________________________________________________*/

/* Starting node of a report structure.
 *      pTitle points at a title for the report defined by the user.
 *      pComment points at a user-defined comment that's to go at the top 
 *          of the report.
 *      pHeader points at the header information in the ODL input file to be displayed
 *      in the report structure of the HTML file.
 *      pLevelOne points at a lstList of reportNodes that were build 
 *          according to the first instruction that was applied.
*/
struct REP
{
  char   *pTitle;
  char   *pComment;
  char   *pHeader;        
  lstLIST 	pLevelOne;
};
typedef struct REP *REPORT;

/* ReportNodes of the report structure representing items that are to be
 *      written in the output of the tool.  Typically there's a reportNode
 *      for each component that is to appear in the output.  Here a component
 *      cound be the name of an ODL group or object and point at other
 *      reportNodes representing the structure of the group/object, or it 
 *      could reprsent a table or a 'parameter = value' relationship.
 *
 *		pName points at the identifier that's to be written in the output.
 *      itemKind indicates the kind of reportNode this is.
 *      pItem points at a related data structure; one representing a
 *          substructure: ODL objects, groups, and parameters are represented
 *          with lstList structures, tables are represented by a reportTable
 *          structure. If the corresponding component is to be ignored, pItem 
 *          will be NULL.
 */
struct reportNode
{
	char 		*pName;
	RN_KIND		itemKind;        
	void 		*pItem; 	
};
typedef struct reportNode *REP_NODE;

/* Tables are represented by the following structure.
 *     ppTable is a one diemnsional array of string pointers.  There are
 *     numRows x numCols such pointers.  The pointers are stored row-wise,
 *     that is, ppTable[1]...ppTable[numCols-1] point at the entries for the 
 *     first row of the table ppTable[numCols]...ppTable[2*numCols-1] point
 *     at the entries for the second, and so on. 
 */

	struct reportTable
	{
 	    char 		*pCaption;
	    unsigned int	numCols;
	    unsigned int	numRows;
	    char		*pColTitle;
	    char		**ppColHeaders;
	    char		*pRowTitle;
	    char		**ppRowHeaders;
	    char		**ppTable;
	};

/* ---------------------------------------------------------------------*/

/* EXTERN PROTOTYPES: the following prototypes are for functions defined
 *     in cmpIFCompiler.c but which are intended for use elsewhere.
*/
int cmpPass1( AGGREGATE pIFTree );
int cmpPass2( struct REP *pReport, OBJECT pRoot );

#endif 
