/* FILE: lstList.h
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION
 *     This is a header file for a collection of data structures
 *     and simple functions that implement a list handling capability.
 *     
 *     Here a lstList is a doubly linked list of arbitrary data 
 *     structures. They are implemented using two typedef'd data 
 *     structures (lstLIST and lstNODE) and a set of special purpose 
 *     functions.  A list will always have a root node that is a struct
 *     lstList and which is pointed to by a lstLIST.  A list will have 0
 *     or more nodes that are lstNode structures pointed at by lstNODE
 *     pointers. 
 *
 * RCS INFO:
 *
 *	$Id: lstList.h,v 1.1 2002/05/08 15:20:20 yin Exp $
 *
 *	$Log: lstList.h,v $
 *	Revision 1.1  2002/05/08 15:20:20  yin
 *	Initial revision
 *
 *	Revision 1.3  1999/04/06 16:16:43  ToolsDev
 *	Promotion of branch 1.2.0 (cleared wlint warnings) for UFM release.
 *
 * Revision 1.2.0.1  1999/02/22  15:32:10  ToolsDev
 * Improved prototype for lstInitList to eliminate warnings.
 *
 * Revision 1.2  1999/02/03  16:32:15  stroup
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.3  1999/01/18  18:33:39  stroup
 * Added prototype for new function lstIsListOfEmptyStrings().
 *
 * Revision 1.1.0.2  1999/01/14  19:36:19  jarva
 * Added prototype for lstDestructor.
 *
 * Revision 1.1.0.1  1999/01/12  20:37:37  stroup
 * Added prototypes for lstJoinList, lstPluckCurrSetNext, & lstPluckCurrSetPrev.  Some minor editing.
 *
 * Revision 1.1  1999/01/06  21:10:54  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
#ifndef _lstList_header
#define _lstList_header

/* INCLUDED FILES: the following included files are here because of
 *     dependencies that exist within this header file.
*/
#include <stdio.h>


/* ---------------------------------------------------------------------*/

/* DATA DECLARATIONS & TYPEDEFS:
 *
*/
#define lstLISTNULL ( (lstLIST) NULL )
#define lstNODENULL ( (lstNODE) NULL )

/* ---------------------------------------------------------------------*/

/* EXTERN VARIABLES: the following variables may be used in other files.
 *    They are defined in lstList.c.
*/
    struct lstNode{
      struct lstNode  *prev;  /* points at previous node in list */
      struct lstNode  *next;  /* points at succeeding node in list */
      void            *item;  /* points at object being listed */
    };

    typedef struct lstNode *lstNODE;

    struct lstList {
      lstNODE first;    /* the first node in the list */
      lstNODE last;     /* the current node in the list */
      lstNODE current;  /* the current node in an iteration */
      int     count;    /* the number of nodes currently in the list */
      char    isA[5];   /* used to mark lstList struct as such */
    };

    typedef struct lstList *lstLIST;


/* ---------------------------------------------------------------------*/

/* EXTERN PROTOTYPES: the following prototypes are for functions defined
 *     in lstList.c but which are intended for use 
 *     elsewhere.
*/
  void lstInitialization( int debugFlag, int verboseFlag, int noExit,
  FILE *pDebugFile, FILE *pUserMessages );

  lstLIST lstInitList( void ); /* Initialize a list */
  lstLIST lstFreeList( lstLIST pList ); /* Free the root node and
                                         * all the list nodes. */
  void  lstAddFirst( lstLIST pList, void *pItem );
  void  lstAddLast( lstLIST pList, void *pItem );
  void *lstPluckFirst( lstLIST pList );
  void *lstPluckLast( lstLIST pList );
  void *lstPluckCurrSetNext (lstLIST pList);
  void *lstPluckCurrSetPrev (lstLIST pList);
  void *lstIterator( lstLIST pList );
  void *lstReverseIterator( lstLIST pList );
  void  lstStopIteration( lstLIST pList );
  int   lstAppendTo( lstLIST pList1, lstLIST pList2 );
  int   lstNumEl( lstLIST pList );
  void *lstReplaceCurrent( lstLIST pList, void *pItem );
  void *lstDestructor(  lstLIST pList, void *(*pDestructor)( void *)  );
 
  /* The following functions are for lists of strings: */
  int      lstIsStrInList( lstLIST pList, char *pStr );
  void     lstPrintStringList( FILE *pF, lstLIST pList, char *pIndent );
  int      lstNumTimesInList( lstLIST pChildList, char *pCompName );
  lstLIST  lstSplitList (char delimiter, char *delimited_string);
  char    *lstJoinList (char *delimiter, lstLIST pList);
  lstLIST  lstMatchList( lstLIST pList, char *indexString );
  int      lstIsListOfEmptyStrings (lstLIST pList);
#endif 
