/* FILE: strUtilities.c
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION: Utility functions for sting manipulation
 *
 * RCS INFO:
 *
 *      $Id: strUtilities.c,v 1.3 1999/04/06 16:19:21 ToolsDev Exp $
 *
 *      $Log: strUtilities.c,v $
 *      Revision 1.3  1999/04/06 16:19:21  ToolsDev
 *      Promotion of branch 1.2.0 (cleared wlint warnings) for UFM release.
 *
 * Revision 1.2.0.1  1999/03/06  17:14:36  ToolsDev
 * Cleared lint warnings.
 *
 * Revision 1.2  1999/02/03  16:35:37  ToolsDev
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.2  1999/01/21  19:20:31  stroup
 * Added strNextToken, strLeftTrim, strRightTrim, and strrev.  Added include
 * of <ctype.h>.
 *
 * Revision 1.1.0.1  1999/01/12  21:43:52  ToolsDev
 * New function strDestructor().
 *
 * Revision 1.1  1999/01/06  21:08:32  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
 * NOTICE: this file was constructed from a template that supplied a
 *         control structure and some simple functions for reporting
 *         errors and debug information. See the static struct 
 *         strCONTROLS below.
*/

/* INCLUDED FILES:
*/
#include "strUtilities.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

/* ---------------------------------------------------------------------*/
/* MACRO DEFINITIONS */

#define VERBOSE_DEBUG 0  /* Set to 1 for maximal debug output but also
                          * see strCONTROLS.debug.

/* ---------------------------------------------------------------------*/
/* GLOBAL VARIABLES: the following variables may be used in other files.
 */
	/* NONE */

/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES: the following variables are defined in other files.
*/
	/* NONE */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
struct strControls {
    /* Controls the message writing functions given in the source file
	 * template (e.g., strDebugMsg().  This structure is
	 * imediately followed by it default settings. Typically the
	 * initailization function strInitialization is invoked by
	 * the calling application to override these settings. */
	int debug;
	int verbose;
	int noExit;
	FILE *pDebugFile;
	FILE *pUserMessages;
} strCONTROLS =
{
	1,
	1,
	0,
	stderr,
	stderr
};

/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static void strFatalError( char *functionName, char *msg );
static void strUserWarning( char *functionName, char *msg );
static void strDebugMsg( char *functionName, char *msg );


/* ---------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: strInitialization
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     debugFlag     -- (int) Control to allow debug print to be output. Set
 *                      to 1 for ouput; 0 otherwise.
 *     verboseFlag   -- (int) Control to allow verbose messages for the user
 *                      to be output. Set to 1 for ouput; 0 otherwise.
 *	   noExit        -- (int) Control to prevent the functions within this
 *                      module from exiting the program.  Set to 1 and all
 *                      functions return to the caller, including 
 *                      strFatalError(...).
 *     pDebugFile    -- (FILE *) Pointer to the debug file to which debug
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 *     pUserMessages -- (FIle *) Pointer to the debug file to which user
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct strCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void strInitialization( int debugFlag, int verboseFlag, int noExit,
	FILE *pDebugFile, FILE *pUserMessages )
{
	strCONTROLS.debug = debugFlag;
	strCONTROLS.verbose = verboseFlag;
	strCONTROLS.noExit = noExit;
	if( pDebugFile ) strCONTROLS.pDebugFile = pDebugFile;
	if( pUserMessages ) strCONTROLS.pUserMessages = pUserMessages;
	return;
} /* end strInitialization */

/*****************************************************************************
 * 
 * NAME: strFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct strCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void strFatalError( char *functionName, char *msg )
{
	if( strCONTROLS.debug )
		fprintf( strCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	if( !strCONTROLS.noExit ) {
		strUserWarning( "strFatalError", 
		     "Fatal programming error is causing the program to shut down." );
		exit(-1);
	}
	else return;

} /* end strFatalError */

/*****************************************************************************
 * 
 * NAME: strUserWarning
 *
 * DESCRIPTION: Reports warnings to the user if the VERBOSE flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct strCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void strUserWarning( char *functionName, char *msg )
{
	if( !strCONTROLS.verbose ) return;

    fprintf( strCONTROLS.pUserMessages, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end strUserWarning */
/*****************************************************************************
 * 
 * NAME: strDebugMsg
 *
 * DESCRIPTION: Reports debug messages for the maintenance programmer if the
 *              debug flag is set and VERBOSE_DEBUG macro is 1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct strCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void strDebugMsg( char *functionName, char *msg )
{
	if( strCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( strCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(strCONTROLS.pDebugFile);
	}
	return;

} /* end strDebugMsg */
 
/*****************************************************************************
 * 
 * NAME: strToUpStrcmp
 *
 * DESCRIPTION: This function performs a case-insensitive comparison of two
 *     strings lexicographically.  First, an uppercase copy of the strings are
 *     made.  Then these string copies are passed to the C standard library
 *     function strcmp for the actual comparison.
 *
 * PARAMETERS:
 *     s1 (const char *)  Pointer to the first input string.
 *     s2 (const char *)  Pointer to the second input string.
 *
 * RETURN VALUE:
 *     strToUpStrcmp (int)   0 --> s1 lexicographically equal to s2;
 *                        1 --> s1 lexicographically greater than s2;
 *                       -1 --> s1 lexicographically less than s2;
 *                       -9 --> s1 or s2 are null; or memory allocation failed
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer either input string is NULL.
 *        POST: Function returns an error indication (-9).
 *     2. PRE:  Pointers to both input strings are not NULL, but insufficient
 *              memory on the heap.
 *        POST: Function returns an error indication (-9).
 *     3. PRE:  Non-null input pointers and sufficient memory on heap.
 *        POST: Function returns 0, 1, or -1.
 *
 * NOTES:
 *     Memory is allocated to store the uppercase versions of s1 and s2.
 *
 *     If either of the input pointers is null, or if a failure occurs
 *     allocating memory, this function returns null.
 *
 * REVISION HISTORY:
 *     10/09/1998  J.D. Stroup  Original
 *     11/19/98    Ken Jarva    Copied from ListExpander and changed name.
 *
 ****************************************************************************/
int strToUpStrcmp (const char *s1, const char *s2)
{
  int  cmp_val;
  unsigned int i, len1, len2;
  char *upper_s1;
  char *upper_s2;

  const int BadDog = -9;

  /* Make sure both input pointers are not NULL. */

  if ((s1 == NULL) || (s2 == NULL)) return (BadDog);

  /* Calculate length of both input strings. */

  len1 =  strlen (s1);
  len2 =  strlen (s2);

  /* Allocate space for the uppercase copies of the given strings. */

  upper_s1 = (char *) malloc ((len1 + 1) * sizeof (char));
  upper_s2 = (char *) malloc ((len2 + 1) * sizeof (char));

  /* Make sure the memory allocations completed successfully. */

  if ((upper_s1 == NULL) || (upper_s2 == NULL)) return (BadDog);

  /* Convert both input strings to uppercase. */

  for (i = 0; i < len1; i++)
    upper_s1[i] = (unsigned char) toupper ( (int) s1[i]);

  for (i = 0; i < len2; i++)
    upper_s2[i] = (char) toupper ( (int) s2[i] );

  /* Put in string terminator. */

  upper_s1[len1] = upper_s2[len2] = '\0';

  /* Compare the uppercase versions of the input strings. */

  cmp_val = strcmp (upper_s1, upper_s2);

  /* Free the memory allocated to the string copies. */

  free (upper_s1);
  free (upper_s2);

  /* Return the comparison results. */

  return (cmp_val);

} /* end strToUpStrcmp */

/*****************************************************************************
 * 
 * NAME: strNextToken
 *
 * DESCRIPTION: This function works similar to the standard library function
 *     "strtok".  It differs in the way it handles "empty" tokens at the
 *     beginning or middle of an input string.  Strtok essentially ignores
 *     them; strNextToken does not.  For example, given the string ":b", where
 *     the delimiter character is the colon (":"), strtok only returns one
 *     token, the "b".  strNextToken returns two tokens: an initial "empty"
 *     token (before the colon) and the "b" that follows the colon.  As
 *     another example, given the string "a::b", strtok returns two tokens,
 *     the "a" and the "b".  strNextToken returns three tokens: the "a", an
 *     "empty" token between the colons, and the "b".  Both functions handle
 *     an "empty" token at the end of an input string the same.  So, given the
 *     input string "a:", both function return only one token, the "a".  (See
 *     NOTES, below, for use of this function.)
 *
 * PARAMETERS:
 *     string (char *)     Pointer to the input string that contains the
 *                         delimited field (tokens).
 *     delimiter (char *)  Pointer to a string of delimiters that this
 *                         function will split the input string on.
 *
 * RETURN VALUE:
 *     strNextToken (char *)  Pointer to a string containing the next token
 *                            retrieved from the input string.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  First parameter is not a NULL pointer, i.e., it points to a
 *              string.
 *        POST: Function saves the input string and returns a pointer to the
 *              first field (token) in the newly saved string.
 *     2. PRE:  First parameter is NULL and this function called previously
 *              under condition (1).
 *        POST: Function returns a pointer to the next field (token) in the
 *              saved string.
 *     3. PRE:  First parameter is NULL and this is the first call to this
 *              function.
 *        POST: Function returns a NULL.  Note that if this is the first call
 *              to this function, there is no saved string yet.
 *     4. PRE:  Insufficient memory on heap (only applicable when the first
 *              parameter is not NULL).
 *        POST: Function returns a NULL pointer.
 *     5. PRE:  Second parameter is NULL.
 *        POST: Function returns a pointer to the remaining saved string.
 *              Strtok, on the other hand, does not handle this situation: it
 *              aborts with a core dump.
 *
 * NOTES:
 *     Refer to the explanation of strtok for a description of how this
 *     function works.  Also see item 4 under "PRE- & POST-CONDITIONS".
 *
 *     Memory is allocated by this function in order to store and save a given
 *     input string.  The input string must be saved for subsequent calls that
 *     wish to retrieve the next token from the string.  Memory is also
 *     allocated to store a copy of the token that this function returns.
 *
 *     If a failure occurs allocating memory, this function returns null.
 *
 * REVISION HISTORY:
 *     09/16/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved NextToken from module ListExpander.c and
 *                              renamed.
 *
 ****************************************************************************/

char *strNextToken (char *string, char *delimiter)
{
  /* The values contained in these pointers can not be lost between
     invocations of this function. Thus, the use of "static". */

  static char *p     = NULL;
  static char *start = NULL;
  static char *token = NULL;

	/* If the first parameter is a string, then free the memory occupied by any
	   old string that this function may have allocated previously, allocate
		 memory for the new string, copy the given input string into the newly
		 allocated memory space, and initialize the pointers. */

  if (string != NULL)
  {
    free (start);
    start = strdup (string);

    /* If error occurred allocating memory, then return NULL. */
		if (start == NULL) return (NULL);

    token = p = start;
  }

	/* Otherwise the first parameter is NULL.  Thus, the calling function wants
	   to retrieve the next token from the saved string.  If there is no saved
     string yet (i.e., this is the first call to strNextToken), return
     NULL. */

  else
  {
    if (p == NULL)
      return (NULL);
    else
      token = p;
  }

	/* If we're at the end of the saved string, then we're done with this
	   string, i.e., all tokens have been identified and returned.  Therefore,
	   return NULL. */

  if (*p == '\0') return (NULL);

	/* If the pointer to the string of delimiters is not NULL, then step through
	   the saved string looking for the next delimiter, or the end-of-string
	   marker. */

	if (delimiter != NULL)
	{
    while ((strchr (delimiter, (int) *p) == NULL) && (*p != '\0')) 
      p++;

  	/* If we encountered a delimiter while stepping through the string, then
	     replace it with the end-of-string marker and bump "p" to point to the
	     beginning of the next token.  Thus we'll be prepared for a subsequent
	     call that wishes to retrieve the next token. */

    if (*p != '\0')
		{
      *p = '\0';
      p++;
		}
	}

	/* Otherwise, the string delimiters pointer is NULL, then step through the
	   saved string until the end-of-string marker is found. */

	else
	{
    while (*p != '\0')
      p++;
	}

	/* Return the next token. */
  {
    char *token_out = strdup (token);

    return (token_out);
  }
} /* end strNextToken */

/*****************************************************************************
 * 
 * NAME: strLeftTrim
 *
 * DESCRIPTION: This function removes all leading white space characters from
 *     the given input string.  A pointer to the trimmed string is returned.
 *
 * PARAMETERS:
 *     s (char *)  Pointer to the input string.
 *
 * RETURN VALUE:
 *     strLeftTrim (char *)  Pointer to a string with no leading white space.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer to input string is NOT NULL.
 *        POST: Function returns a pointer to a string that is
 *              identical to the input string, but without leading white
 *              space.
 *     2. PRE:  Pointer to input string is NULL.
 *        POST: Function returns a NULL pointer.
 *
 * NOTES:
 *     No memory is allocated for the trimmed string.  Thus, the returned
 *     pointer points to a substring of the input string.  If there is no
 *     leading white space, the returned pointer points to the same string
 *     that the input string points to.
 *
 *     If the pointer to the input string is null, this function returns null.
 *
 * REVISION HISTORY:
 *     09/15/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved ltrim from module ListExpander.c and
 *                              renamed.
 *
 ****************************************************************************/

char *strLeftTrim (char *s)
{
	/* If the pointer to the input string is not NULL, then step through the
	   input string character by character until a non-white space character (or
	   the end-of-string marker) is encountered.  Return a pointer to this first
	   non-white space character. Note that if the input pointer is NULL, a NULL
	   pointer (i.e., the input pointer) is returned. */

  if (s != NULL)
  {
    while (  isspace ( (int) *s ) && ( *s != '\0' )  )
      s++;
  }

  return (s);
} /* end strLeftTrim */

/*****************************************************************************
 * 
 * NAME: strRightTrim
 *
 * DESCRIPTION: This function removes all trailing white space characters from
 *     the given input string.  A pointer to the trimmed string is returned.
 *
 * PARAMETERS:
 *     s (char *)  Pointer to the input string.
 *
 * RETURN VALUE:
 *     strRightTrim (char *)  Pointer to a string with no trailing white
 *                            space.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer to input string is NOT NULL and there is sufficient
 *              memory on heap.
 *        POST: Function returns a pointer to a string that is that is
 *              identical to the input string, but without trailing white
 *              space.
 *     2. PRE:  Pointer to input string is NULL.
 *        POST: Function returns a NULL pointer.
 *     3. PRE:  Insufficient memory on heap.
 *        POST: Function returns a NULL pointer.
 *
 * NOTES:
 *     Memory is allocated for the trimmed string via the function strrev.
 *     Thus, the returned pointer points to a COPY of a substring of the input
 *     string.  If there is no trailing white space, the returned pointer
 *     points to an exact copy of the input string.
 *
 *     If the pointer to the input string is null, or if a failure occurs
 *     allocating memory, this function returns null.  In both cases no memory
 *     has been allocated.
 *
 * REVISION HISTORY:
 *     09/15/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved rtrim from module ListExpander.c and
 *                              renamed.
 *
 ****************************************************************************/

char *strRightTrim (char *s)
{
  char *rev, *new_s;

  if (s == NULL)    /* pointer to the input string is NULL; therefore... */
    return (NULL);  /* ... return NULL */

	/* Reverse the input string.  Trailing white space, if any, becomes LEADING
	   white space. */

  rev = strrev (s);

	/* If the previous call to strrev returned a NULL pointer, then strrev
	   failed to allocate memory. In this case, this function will return a
	   NULL pointer (see the "else" clause, below). */

  if (rev != NULL) /* strrev successfully allocated memory */
	{
  	/* Trim the leading white space of the reversed input string.  Reverse the
		   string again, putting it back into its original order. */

    new_s = strrev (strLeftTrim (rev));

	  /* Free the memory allocated to the reversed string. */

		free (rev);
	}

	else
		new_s = NULL;

	/* Return a pointer to the new string that contains no trailing white
	   space. */

  return (new_s);
} /* end strRightTrim */

/*****************************************************************************
 * 
 * NAME: strrev
 *
 * DESCRIPTION: This function reverses the given input string.  A pointer to
 *     the reversed string is returned.
 *
 * PARAMETERS:
 *     s (char *)  Pointer to the input string.
 *
 * RETURN VALUE:
 *     strrev (char *)  Pointer to a string that is the reverse of s.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer to input string is NOT NULL and there is sufficient
 *              memory on heap.
 *        POST: Function returns a pointer to a string that is the reverse of
 *              the input string.
 *     2. PRE:  Pointer to input string is NULL.
 *        POST: Function returns a NULL pointer.
 *     3. PRE:  Insufficient memory on heap.
 *        POST: Function returns a NULL pointer.
 *
 * NOTES:
 *     Memory is allocated for the reversed string.
 *
 *     If the pointer to the input string is null, or if a failure occurs
 *     allocating memory, this function returns null.
 *
 * REVISION HISTORY:
 *     09/16/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved from module ListExpander.c.
 *
 ****************************************************************************/

char *strrev (char *s)
{
  char *t;

	/* If the pointer to the input string is NOT NULL */

  if (s != NULL)
  {
    unsigned int len = strlen (s); /* determine length of input string */

    /* Allocate memory for the new, reversed string. (REMEMBER: add 1 space
       for end-of-string marker.) */

    t = (char *) malloc (sizeof (char) * (len + 1));

		if (t != NULL) /* memory allocation succeeded */
		{
      int i;
      char *r = s + len - 1; /* r points to end of string s */

			/* Step backwards through the input string, using r, and assigning
         characters to the newly allocated string. */

      for (i = 0; r >= s; r--)
         t[i++] = *r;
    
      t[len] = '\0'; /* don't forget the end-of-string marker */
		}
  }
	else /* the pointer to the input string was NULL: return a NULL pointer */
		t = s;

	/* Return the pointer to the new, reversed string, or NULL. */

  return (t);
} /* end strrev */

/*****************************************************************************
 * 
 * NAME: strDestructor
 *
 * DESCRIPTION: Frees the given string.
 *
 * PARAMETERS:  
 *     pStr -- (void *) Pointer to a string. See note 1 below.
 *
 * RETURN VALUE: (void *) NULL
 *     
 * GLOBAL VARIABLES: None
 *      
 * NOTES:
 *     1. The function's formal argument is a void pointer rather than
 *        a character pointer because this function is frequently used
 *        as an argument to lstDestructor, which must apply strDestructor
 *        without having knowledge of the data type of strDestructor's
 *        argument.
 *     2. This function always returns a (void *) NULL.  This is to accomodate
 *        the following usage:
 *                             pStr = strDestructor( pStr );
 *        whereby the string to which pStr points is freed and pStr is set
 *        to the NULL pointer in a single statement.
 *
 * REVISION HISTORY:
 *   01/12/99  Ken Jarva  --  Original
 *
 ****************************************************************************/

void *strDestructor(void *pStr)
{
        if( pStr )free( pStr );
        return (void *)NULL;
} /* end strDestructor */


