/* FILE: synODL.c
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION: Functions that build and use data dictionaries for the 
 *              puprose of syntax checking.
 *
 * RCS INFO:
 *
 *      $Id: synODL.c,v 1.1 2002/05/08 15:20:20 yin Exp $
 *
 *      $Log: synODL.c,v $
 *      Revision 1.1  2002/05/08 15:20:20  yin
 *      Initial revision
 *
 *      Revision 1.3  2000/04/13 00:04:47  ToolsDev
 *      *** empty log message ***
 *
 * Revision 1.3  1999/04/06  16:26:41  ToolsDev
 * Promotion of branch 1.2.0 (cleared wlint warnings) for UFM release.
 *
 * Revision 1.2.0.3  1999/03/06  15:45:10  ToolsDev
 * More mods for lint.
 *
 * Revision 1.2.0.2  1999/02/22  15:47:17  ToolsDev
 * Changed prototypes to eliminate warnings.
 *
 * Revision 1.2.0.1  1999/02/18  21:20:36  ToolsDev
 * Corrections to satify gcc's -Wall; added headers; and finalized
 * a couple of functions that'd been left as stubs (memory management).
 *
 * Revision 1.2  1999/02/03  16:29:28  stroup
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.1  1999/01/20  19:33:29  stroup
 * Replaced ODL_Node.h include with odlSupport.h include.
 *
 * Revision 1.1  1999/01/06  21:24:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
 * NOTICE: this file was constructed from a template that supplied a
 *         control structure and some simple functions for reporting
 *         errors and debug information. See the static struct 
 *         synCONTROLS below.
*/

/* INCLUDED FILES:
*/
#include "synODL.h"
#include "odlSupport.h"
#include "strUtilities.h"

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

/* ---------------------------------------------------------------------*/
/* MACRO DEFINITIONS */

#define VERBOSE_DEBUG 1  /* Set to 1 for maximum debug output but also
                          * see synCONTROLS.debug. */

/* ---------------------------------------------------------------------*/
/* GLOBAL VARIABLES: the following variables may be used in other files.
 */
	/* NONE */

/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES: the following variables are defined in other files.
*/
	/* NONE */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
struct synControls {
    /* Controls the message writing functions given in the source file
	 * template (e.g., synDebugMsg().  This structure is
	 * imediately followed by its default settings. Typically the
	 * initailization function synInitialization is invoked by
	 * the calling application to override these settings. */
	int debug;
	int verbose;
	int noExit;
	FILE *pDebugFile;
	FILE *pUserMessages;
} synCONTROLS =
{
	1,
	1,
	0,
	NULL,
	NULL
};

/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static void synFatalError( char *functionName, char *msg );
static void synUserWarning( char *functionName, char *msg );
static void synDebugMsg( char *functionName, char *msg );
static synAGGDEF synInitAggDef( void );
static synPARDEF synInitParDef( void );
static synAGGDEF synBuildAggDef( AGGREGATE pODLDef );
static synPARDEF synBuildParDef( AGGREGATE pODLDef );
static int synDisambiguate( lstLIST pList, sltLUT pAmbiguous );
/* ---------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: synInitialization
 *
 * DESCRIPTION: Initializes the utility's control structure, overriding the 
 *              established defaults. (See synControls struct above.)
 *
 * PARAMETERS:  
 *     debugFlag     -- (int) Control to allow debug print to be output. Set
 *                      to 1 for ouput; 0 otherwise.
 *     verboseFlag   -- (int) Control to allow verbose messages for the user
 *                      to be output. Set to 1 for ouput; 0 otherwise.
 *	   noExit        -- (int) Control to prevent the functions within this
 *                      module from exiting the program.  Set to 1 and all
 *                      functions return to the caller, including 
 *                      synFatalError(...).
 *     pDebugFile    -- (FILE *) Pointer to the debug file to which debug
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 *     pUserMessages -- (FIle *) Pointer to the debug file to which user
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct synCONTROLS:
 *     pDebugFile -- (FILE *) The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *     pUserMessages -- (FILE *) The file to which user messages are written.
 *     verbose    -- (int) control allowing verbose messages to be written
 *                   to the user.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void synInitialization( int debugFlag, int verboseFlag, int noExit,
	FILE *pDebugFile, FILE *pUserMessages )
{
	synCONTROLS.debug = debugFlag;
	synCONTROLS.verbose = verboseFlag;
	synCONTROLS.noExit = noExit;
	if( pDebugFile ) synCONTROLS.pDebugFile = pDebugFile;
	if( pUserMessages ) synCONTROLS.pUserMessages = pUserMessages;
	return;
} /* end synInitialization */

/*****************************************************************************
 * 
 * NAME: synFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct synCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void synFatalError( char *functionName, char *msg )
{
	if( synCONTROLS.debug )
		fprintf( synCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	if( !synCONTROLS.noExit ) {
		synUserWarning( "synFatalError", 
		     "Fatal programming error is causing the program to shut down." );
		exit(-1);
	}
	else return;

} /* end synFatalError */

/*****************************************************************************
 * 
 * NAME: synUserWarning
 *
 * DESCRIPTION: Reports warnings to the user if the verbose flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct synCONTROLS:
 *     pUserMessages -- FILE The debug file to which error messages are to be 
 *                      written.
 *     verbose       -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void synUserWarning( char *functionName, char *msg )
{
	if( !synCONTROLS.verbose ) return;

    fprintf( synCONTROLS.pUserMessages, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end synUserWarning */
/*****************************************************************************
 * 
 * NAME: synDebugMsg
 *
 * DESCRIPTION: Reports debug messages for the maintenance programmer if the
 *              debug flag is set and VERBOSE_DEBUG macro is 1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct synCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void synDebugMsg( char *functionName, char *msg )
{
	if( synCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( synCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(synCONTROLS.pDebugFile);
	}
	return;

} /* end synDebugMsg */
 
/*****************************************************************************
 * 
 * NAME: synInitDataDict()
 *
 * DESCRIPTION: Allocates space for and initializes a data dictionary.  This
 *      function must be called before a DD can be built or used.
 *
 * PARAMETERS:  None
 *
 * RETURN VALUE:
 *     synInitDataDict -- (synDD) Pointer to data dictionary of definitions
 *                                of ODL components.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. In this implementation, a data dictionary is a set of lookup tables
 *        (one for each letter of the alphabet and one for everything else)
 *        indexed on the name of the thing being defined. The entry of the
 *        LUT is a structure containing an indicator of the kind of thing 
 *        being defined and a pointer to the definition.
 *
 * REVISION HISTORY:
 *   10/23/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
synDD synInitDataDict(void)
{
	synDD pDD;
	int i;

	pDD = malloc( sizeof( *pDD ) );
	for( i = 0; i < 27; i++ )
		pDD->section[i] = sltInitLUT();

	return pDD;
	
} /* end synInitDataDict */
/*****************************************************************************
 * 
 * NAME: synAddToDD
 *
 * DESCRIPTION: Adds a given ODL component definition to the given data 
 *		dictionary
 *
 * PARAMETERS:  
 *     pDD    -- (synDD) Pointer to the data dictionary being manipulated.
 *     pName  -- (char *) Pointer to the name of the thing being defined.
 *     pEntry -- (synENTRYDD *) Pointer to the entry structure (kind
 *                              indicator an pointer to def).
 *
 * RETURN VALUE:
 *     synAddToDD -- (int) Indicator of success; 0 -> failure; 1 -> success;
 *                         2 -> multiply defined. The function asserts that
 *                         none of the given pointers is NULL.
 *     
 * GLOBAL VARIABLES:
 *     <<variable name>> -- <<TYPE>> <<Description>>
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   10/23/98  Ken Jarva  --  Original
 *   12/08/98  Ken Jarva  --  Major revision to improve lookup speed.
 *
 ****************************************************************************/
int synAddToDD(synDD pDD, char *pName, synENTRYDD pEntry )
{
	char *letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int i;

	if( !pDD || !pName || !pEntry ) return 0;
	else {
		synENTRYDD pOriginal; /* Will be nonNULL if the DD already
							   * contained a definition for this term */
		int firstLetter; /* First letter of the term being defined */

		/* Determine which section of the DD to use: */
		firstLetter =  toupper( (int) *pName );
		for(i = 0; firstLetter != letters[i]; i++ );
		if( i >= 26 ) i = 26;

		/* Insert the definition into the dictionary: */
		pOriginal = 
			(synENTRYDD) sltAddOrReplace( (sltLUT) pDD->section[i], 
			pName, (void *) pEntry );

		/* If an original definition is being replaced, free the space used
		 * by the old one. */
		if( pOriginal ) {
			if( pOriginal->kind == synAGGREGATE ) 
				                             synFreeAggDef( pOriginal->pDef );
			else synFreeParDef( pOriginal->pDef );
			free( pOriginal );
			return 2; /* multiply defined item */
		}
		return 1; /* normal return */
	}
} /* end synAddToDD */

synENTRYDD synDDLookUp( synDD pDD, char *pIndex )
{
	char *letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int i, firstLetter;

	/* Determine which section of the DD to use: */
	firstLetter = toupper( (int) *pIndex );
	for(i = 0; firstLetter != letters[i]; i++ );
		if( i >= 26 ) i = 26;

	return sltLookUp( pDD->section[i], pIndex );
}

/*****************************************************************************
 * 
 * NAME: synBuildAggDef
 *
 * DESCRIPTION: Converts an AggregatDescriptor from a data dictionary into
 *		the internal format (struct synAggDef).
 *
 * PARAMETERS:  
 *     pODLDef -- AGGREGATE Pointer to a object node in the data dictionary
 *			                tree containing a definition of an ODL aggregate.
 *
 * RETURN VALUE:
 *     synBuildAggDef -- (synAGGDEF) Pointer to the new aggregate definition
 *			                         structure.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     1. If the ODL from the DD isn't pretty much as expected, this function
 *        invokes fatal error processing.
 *     2. It just ignores objects that are not selection lists.
 *
 * REVISION HISTORY:
 *   10/26/98  Ken Jarva  --  Original
 *   02/18/99  Ken Jarva  --  strdup'd pAggName so the internal definition 
 *                            will be independent of the DD tree.
 *
 ****************************************************************************/
synAGGDEF synBuildAggDef( AGGREGATE pODLDef )
{
	PARAMETER pParm;
	synAGGDEF pDef;
	AGGREGATE pObj;
	lstLIST   pList;
	struct ODL_Node *pNode;
	int unambiguous = 1;

	pDef = synInitAggDef(); /* allocate & initialize an aggregate definition
							  * structure */

	/* The Name of the thing being defined must be given... */
	pParm = FindParameter( pODLDef, "NAME" );
	if( pParm ) pDef->pAggName = 
                              strdup( pParm->first_value->item.value.string );
	else {
		synFatalError("synBuildAggDef",
			"Bad data dictionary. No NAME in agg. def." );
	}

	/* The Kind of the thing being defined must be given... */
	pParm = FindParameter( pODLDef, "KIND" );
	if( pParm ) {
		char *pKind;
		pKind = pParm->first_value->item.value.string;
		if( strToUpStrcmp( pKind, "OBJECT" ) == 0 ) pDef->aggKind = KA_OBJECT;
		else if( strToUpStrcmp( pKind, "GROUP" ) == 0 ) pDef->aggKind = KA_GROUP;
		else synFatalError( "synBuildAggDef", "definition for unknown agg. kind.");
	} else {
		synFatalError("synBuildAggDef",
			"Bad data dictionary. No KIND in agg. def." );
	}

	/* Process the ExactlyOne list if it exists: */
	  /* Get a pointer to the ODL parameter & convert to node pointer: */
	  pParm = FindParameter( pODLDef, "EXACTLYONE" );
	  /* If there was one, extract the list: */
	  if( pParm ){
		  pNode = odlMakeNode( PARAMETER_NODE, pParm );
		  pList = odlGetValueList( pNode );
	  }
	  else pList = lstInitList(); /* empty list */
	  pDef->pExactlyOne = pList;

	  /* Check for an ambiguous definition and add terms as necessary to
	   * the disambiuation LUT: */
	  if( synDisambiguate( pList, pDef->pAmbiguous ) != unambiguous )
		  pDef->ambiguousDef = 1;

	/* Process the ZeroOrMore list if it exists: */
	  /* Get a pointer to the ODL parameter & convert to node pointer: */
	  pParm = FindParameter( pODLDef, "ZEROORMORE" );
	  /* If there was one, extract the list: */
	  /* If there was one, extract the list: */
	  if( pParm ){
		  pNode = odlMakeNode( PARAMETER_NODE, pParm );
		  pList = odlGetValueList( pNode );
	  }
	  else pList = lstInitList();
	  pDef->pZeroOrMore = pList;

	  /* Check for an ambiguous definition and add terms as necessary to
	   * the disambiuation LUT: */
	  if( synDisambiguate( pList, pDef->pAmbiguous ) != unambiguous )
		  pDef->ambiguousDef = 1;

	/* Process the OneOrMore list if it exists: */
	  /* Get a pointer to the ODL parameter & convert to node pointer: */
	  pParm = FindParameter( pODLDef, "ONEORMORE" );
	  /* If there was one, extract the list: */
	  if( pParm ){
		  pNode = odlMakeNode( PARAMETER_NODE, pParm );
		  pList = odlGetValueList( pNode );
	  }
	  else pList = lstInitList();
	  pDef->pOneOrMore = pList;

	  /* Check for an ambiguous definition and add terms as necessary to
	   * the disambiuation LUT: */
	  if( synDisambiguate( pList, pDef->pAmbiguous ) != unambiguous )
		  pDef->ambiguousDef = 1;

	/* Process the ZeroOrOne list if it exists: */
	  /* Get a pointer to the ODL parameter & convert to node pointer: */
	  pParm = FindParameter( pODLDef, "ZEROORONE" );	  
	  /* If there was one, extract the list: */
	  if( pParm ) {
		  pNode = odlMakeNode( PARAMETER_NODE, pParm );
		  pList = odlGetValueList( pNode );
	  }
	  else pList = lstInitList();
	  pDef->pZeroOrOne = pList;

	  /* Check for an ambiguous definition and add terms as necessary to
	   * the disambiuation LUT: */
	  if( synDisambiguate( pList, pDef->pAmbiguous ) != unambiguous )
		  pDef->ambiguousDef = 1;

	/* Process the select lists if any: */
	  pObj = pODLDef->first_child;
	  while( pObj ) {
		if( strcmp(pObj->name, "SELECTIONLIST") == 0 ) {
			int required = 0; /* indicator of whether the selection list is or
							   * is not requried. */
			pParm = FindParameter( pObj, "REQUIRED" );
			if( pParm ) {
				if( strToUpStrcmp(pParm->first_value->item.value.string, "YES") == 0 )
					required = 1;
				else required = 0;
			} else {
				synFatalError( "synBuildAggDef",
					"DD:AggregateDescriptor/SelectionList/Required "
					"parameter is missing.");
			}
			pParm = FindParameter( pObj, "SELECTONE" );
			if( pParm ) {
				pNode = odlMakeNode( PARAMETER_NODE, pParm );
				pList = odlGetValueList( pNode );
			} else {
				synFatalError( "synBuildAggDef",
					"DD:AggregateDescriptor/SelectionList/SelectOne "
					"parameter is missing.");
			}
			if( required )
				lstAddLast( pDef->pReqSelectLists, (void *)pList );
			else lstAddLast( pDef->pOptSelectLists, (void *)pList );

			/* Check for an ambiguous definition and add terms as necessary to
			 * the disambiuation LUT: */
			if( synDisambiguate( pList, pDef->pAmbiguous ) != unambiguous )
				pDef->ambiguousDef = 1;

		} /* ignore any obects that are not SelectionLists */
		pObj = pObj->right_sibling;
	  }
	  if( pDef->ambiguousDef == 1 ) {
		  char msg[256];
		  sprintf( msg, "Ambiguous definition for %s.", pDef->pAggName );
		  synUserWarning( "synBuildAggDef", msg );
	  }
	  return pDef;

} /* end synBuildAggDef */

/*****************************************************************************
 * 
 * NAME: synDisambiguate
 *
 * DESCRIPTION: This function searches a list of identifiers such as would
 *		be expected to occur in an aggregate definition looking for elements
 *		that contain a discriminator (~n, n an integer). An entry is placed
 *      in the given LUT (pAmbiguous) for each such identifier found. The
 *		index for the entry in the LUT is the ambiguous name (e.g., DATASET)
 *      and the corresponding entry is the unambiguous name (e.g., DATASET~2).
 *      If an entry of in the LUT is changed, the corresponding definition
 *      is considered ambiguous and a zero is returned. This would happen, for
 *      example if a definition contained references to both DATASET~1 and to
 *		DATASET~2.
 *
 * PARAMETERS:  
 *     pList	  -- (lstLIST)  A lstList of unambiguous names such as would
 *                              appear in the definition of an aggregate.
 *	   pAmbiguous -- (sltLUT)   A LUT that allows ambiguous names to be
 *								looked up in order to find the unambiguous
 *								term to which it refers.
 *
 * RETURN VALUE:
 *     synDisambiguate -- (int) 1 -> the list was checked and the LUT was
 *								updated; no ambiguities were discovered.
 *                              0 -> the definition being checked is ambig-
 *								uous. 
 *
 * NOTES:
 *     1. pTemp is used to protect the idenitifier in the given pList (i.e.,
 *        note the use of strdup.)
 *     2. Notice that it is not necessary to allocate space for the 
 *        index string being passed to sltAddOrReplace so using pTemp in this
 *        context is OK.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   12/01/98  Ken Jarva  --  Added header and corrected iteration loop.
 *
 ****************************************************************************/
int synDisambiguate( lstLIST pList, sltLUT pAmbiguous )
{
	char *pName;
	char *pOld;
	int ambiguous = 0, unambiguous = 1, iReturn;

	iReturn = unambiguous;

	lstStopIteration( pList );
	pName = lstIterator( pList );
	while( pName )
	{
		char *pTemp = strdup( pName );
		unsigned int i;
		for( i = 0; i < strlen(pTemp) && pTemp[i] != '~'; i++ );
		if( pTemp[i] == '~' ){
			pTemp[i] = '\0';
			pOld = sltAddOrReplace(pAmbiguous, pTemp, pName);
			if( pOld && strToUpStrcmp(pName, pOld ) != 0 ) {
				char msg[256];
				sprintf( msg, "ambiguous definition: %s and %s\n\twere used"
					" in the same definition.", pName, pOld );
				synUserWarning( "synDisambiguate", msg );
				iReturn = ambiguous;
			}
		}
		free(pTemp);
		pName = lstIterator( pList );
	}
	lstStopIteration( pList );

	return iReturn;

} /* end synDisambiguate */


/*****************************************************************************
 * 
 * NAME: synBuildDataDict
 *
 * DESCRIPTION: Processes the definitions given in an ODL tree and places
 *		them in the given data dictionary.
 *
 * PARAMETERS:  
 *     pDD     -- (synDD)     Initialzed data dictionary structure.
 *     pDDTree -- (AGGREGATE) ODL tree containing definitions to be converted
 *			                  to internal format and added to the DD.
 *
 * RETURN VALUE:
 *     synBuildDataDict -- (int) Success indicator: 1 -> success; 0 -> failure.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES: None
 *
 * REVISION HISTORY:
 *   10/27/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
int synBuildDataDict( synDD pDD, AGGREGATE pDDTree )
{
	int Success = 1, Failure = 0;
	AGGREGATE pObj;      /* points at the ODL definition of interest */
	synAGGDEF pAggDef;   /* points at the definition of an aggregate that's
						  * been processed from the original ODL. */
	synPARDEF pParDef;   /* points at the definition of an parameter that's
						  * been processed from the original ODL. */
	synENTRYDD pDDEntry; /* points at struct used in the DD; it just 
						  * indicates whether the def is for an aggregate or
						  * a parameter, and it gives a pointer to the def. */
	char *pIndex = NULL; /* points at the name of the thing being defined. */

	/* Check the inputs: */
	assert( pDD );
	assert( pDDTree );

	/* Step through all the definitions given in the DD tree: */
	pObj = pDDTree->first_child;
	while( pObj ) {
		if( strcmp(pObj->name, "AGGREGATEDESCRIPTOR") == 0 ) {
			/* This def is for an ODL Aggregate... */

			/* Convert the ODL to the internal reqpresentation: */
			pAggDef = synBuildAggDef( pObj );
			pIndex = pAggDef->pAggName; /* index into the DD */

			/* Allocate space for the DD entry:*/
			pDDEntry = (synENTRYDD) malloc( sizeof(*pDDEntry) );
			assert( pDDEntry );

			/* Set the entries values: */
			pDDEntry->kind = synAGGREGATE;
			pDDEntry->pDef = (void *) pAggDef;

		} else if( strcmp(pObj->name, "PARAMETERDESCRIPTOR") == 0 ) {
			/* This def is for an ODL Parameter... */

			/* Convert the ODL to the internal reqpresentation: */
			pParDef = synBuildParDef( pObj );
			pIndex = pParDef->pName; /* index into the DD */

			/* Allocate space for the DD entry:*/
			pDDEntry = (synENTRYDD) malloc( sizeof(*pDDEntry) );
			assert( pDDEntry );

			/* Set the entries values: */
			pDDEntry->kind = synPARAMETER;
			pDDEntry->pDef = (void *) pParDef;

		}

		/* Add entry to the dictionary: */
		if( !synAddToDD(pDD, pIndex, pDDEntry) ){
			synDebugMsg( "synBuildDataDict", "Fails to add definition to "
				         "data dictionary." );
			return Failure;
		}

		pObj = pObj->right_sibling; /* Next definition */
	} 

	return Success;

} /* end synBuildDataDict */
/*****************************************************************************
 * 
 * NAME: synPrintAggDef
 *
 * DESCRIPTION: Will ouput an aggregate definition to the given file (*pF)
 *
 * PARAMETERS:  
 *     pF   -- (FILE *) File for output.
 *     pDef -- (synAGGDEF) Definition strcutre for an aggregate.
 *
 * RETURN VALUE: None.
 *
 * REVISION HISTORY:
 *   10/27/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
void synPrintAggDef( FILE *pF, synAGGDEF pDef )
{

	lstLIST pList;
	AGGREGATE_KIND kind;

	assert( pF );
	if( !pDef ){
		fprintf( pF,"\nUndefined aggregate!!!\n" );
		return;
	}

	fprintf( pF,"\nDefinition for %s\n", pDef->pAggName );

	/* Aggregate kind...*/
	kind = pDef->aggKind;
	if( kind == KA_OBJECT ) fprintf( pF, "  Agg kind = OBJECT\n" );
	else fprintf( pF, "  Agg kind = GROUP\n" );

	/* ExactlyOne...*/
	pList = pDef->pExactlyOne;
	if( lstNumEl(pList) > 0 ) {
		fprintf( pF,"  The ExactlyOne list: \n" );
		lstPrintStringList( pF, pList, "\t" );
	}

	/* OneOrMore...*/
	pList = pDef->pOneOrMore;
	if( lstNumEl(pList) > 0 ) {
		fprintf( pF,"  The OneOrMore list: \n" );
		lstPrintStringList( pF, pList, "\t" );
	}

	/* ZeroOrMore...*/
	pList = pDef->pZeroOrMore;
	if( lstNumEl(pList) > 0 ) {
		fprintf( pF,"  The ZeroOrMore list: \n" );
		lstPrintStringList( pF, pList, "\t" );
	}

	/* ZeroOrOne...*/
	pList = pDef->pZeroOrOne;
	if( lstNumEl(pList) > 0 ) {
		fprintf( pF,"  The ZeroOrOne list: \n" );
		lstPrintStringList( pF, pList, "\t" );
	}

	/* List all the REQUIRED select lists */
	pList = pDef->pReqSelectLists; /* List of select lists */
	if( lstNumEl(pList) > 0 ) {
		lstLIST pSelectList;
		lstStopIteration( pList );
		  pSelectList = lstIterator( pList );
		  while( pSelectList ) {
			/* Print the current select list: */
			fprintf( pF,"  The Required Select list: \n" );
			lstPrintStringList( pF, pSelectList, "\t" );
			/* Get the next select list */
			pSelectList = lstIterator( pList );
		  }
		lstStopIteration( pList );
	}

	/* List all the OPTIONAL select lists */
	pList = pDef->pOptSelectLists; /* List of select lists */
	if( lstNumEl(pList) > 0 ) {
		lstLIST pSelectList;
		lstStopIteration( pList );
		  pSelectList = lstIterator( pList );
		  while( pSelectList ) {
			/* Print the current select list: */
			fprintf( pF,"  The Required Select list: \n" );
			lstPrintStringList( pF, pSelectList, "\t" );
			/* Get the next select list */
			pSelectList = lstIterator( pList );
		  }
		lstStopIteration( pList );
	}

	fflush(synCONTROLS.pDebugFile);

	return;

} /* end synPrintAggDef */

/*****************************************************************************
 * 
 * NAME: synInitAggResult
 *
 * DESCRIPTION: Allocates and initializes a structure for reporting the 
 *			    results of a syntax check on an ODL aggregate.
 *
 * PARAMETERS:  None
 *
 * RETURN VALUE:
 *     synInitAggResult -- (synAGGRESULT) Pointer to a new, initialized report
 *                                      structure.
 *     
 * GLOBAL VARIABLES: None
 *
 * REVISION HISTORY:
 *   10/28/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
synAGGRESULT synInitAggResult( void )
{
	synAGGRESULT pResult;

	pResult = malloc( sizeof(*pResult) );
	pResult->status = synUNDEFINED;
	pResult->ambiguous = 0;
	pResult->badChildren = 0;
	pResult->wrongKind = 0;
	pResult->pAggName = strdup("");
	pResult->pMissingComponents = lstInitList();
	pResult->pTooManyOccurrences = lstInitList();	
	pResult->pUnknownAggregates = lstInitList();
	pResult->pUnknownParameters = lstInitList();
	pResult->pUnsatisfiedSelections = lstInitList();
	pResult->pChildrenReports = lstInitList();
	pResult->pParameterReports = lstInitList();

	return pResult;

} /* end synInitAggResult */
/*****************************************************************************
 * 
 * NAME: synInitParResult
 *
 * DESCRIPTION: Allocate & initialize the structure used to report the results
 *              of a syntax check on an ODL parameter.
 *
 * PARAMETERS:  None
  *
 * RETURN VALUE:
 *     synInitParResult -- (synPARRESULT) Points at new, initialized report
 *                                        structure.
 *     
 * GLOBAL VARIABLES: None
 *
 * REVISION HISTORY:
 *   10/28/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
synPARRESULT synInitParResult( void )
{
	synPARRESULT pResult;

	pResult = malloc( sizeof(*pResult) );

	pResult->pName = strdup("");
	pResult->badType = 0;
	pResult->badKind = 0;
	pResult->outOfRange = 0;
	pResult->status = synUNDEFINED;

	return pResult;

} /* end synInitParResult */

/*****************************************************************************
 * 
 * NAME: synGetAggDef
 *
 * DESCRIPTION: Looks up the given term (or its disambiguated version)
 *              in the given data dictionary.  This function will only return 
 *              definitions for ODL aggregates. See synGetParDef for the 
 *              definitions of ODL parameters.
 *
 * PARAMETERS:  
 *     pDD    -- (synDD) The dictionary containing the desired definition.
 *     pIndex -- (char *) The term for which a definition is required.
 *	   pParentDef -- (synAGGDEF) Definition of the parent aggregate. It's
 *                used to disambiguate identifiers.
 *
 * RETURN VALUE:
 *     synGetAggDef -- (synAGGDEF) Pointer to the desired definition or NULL.
 *     
 * GLOBAL VARIABLES: None.
 *
 * NOTES:
 *   1. More than one definition for an ODL identifier may exist in the DD.
 *      If so, the various definitions will be listed under identifiers formed 
 *      by adding ~n (where n is an integer) to the original (ambiguous)
 *      identifier.
 *   2. The ~n discriminators only appear in the DD, never in the actuall ODL.
 *   3. When a DD entry is built for an aggregate, the names of any of its 
 *      components containing ~n discriminators are put into a lookup table 
 *      indexed by the ambiguous part of the name. (E.g., MEDIA_TYPE~1 would be
 *      indexed by MEDIA_TYPE.) 
 *   4. A definition that causes two entries for the same index is considered
 *      ambiguous and cannot be handled. 
 *   5. Only the names of ambiguous componentidentifiers are included in the 
 *      LUT, so a failure to find an entry just means that the identifier is 
 *      not ambiguous.
 *
 * REVISION HISTORY:
 *   10/27/98  Ken Jarva  --  Original
 *   11/23/98  Ken Jarva  --  Revised to handle ambiguity.
 *   12/03/98  Ken Jarva  --  Corrected bug: wasn't handling a DD with an
 *                            entry having a ~1 but no ~2.
 *   12/03/98  Ken Jarva  --  Simplified using pParentDef->pAmbiguous.
 *
 ****************************************************************************/
synAGGDEF synGetAggDef( synDD pDD, char *pIndex, synAGGDEF pParentDef )
{
	synENTRYDD pEntry;
	synAGGDEF pReturn = NULL;
	char *pUnambiguousName; /* Name with the definition discrimiator (~n) 
	                         * if present */

	/* Make sure the arguments point at something: */
	assert( pDD );
	assert( pIndex );

	/* Determine the unambiguous name if possible */
	if( pParentDef ) {
		/* A definition for the parent aggregate exists, so... */
		pUnambiguousName = sltLookUp( pParentDef->pAmbiguous, pIndex );
		if( !pUnambiguousName ) pUnambiguousName = pIndex;
	} else
		/* No such definition exists, so assume *pIndex is unambiguous: */
		pUnambiguousName = pIndex;

	/* Now get the definition from the data dictionary: */
	pEntry = synDDLookUp( pDD, pUnambiguousName );

	/* If we got one, return the pointer; otherwise return NULL */
	if( pEntry && pEntry->kind == synAGGREGATE ){
		pReturn = pEntry->pDef; /* The definition from the DD */
	} else pReturn = NULL;

	return pReturn;
		
} /* end synGetAggDef */


/*****************************************************************************
 * 
 * NAME: synGetParDef
 *
 * DESCRIPTION: Looks up the given term (or its disambiguated version)
 *              in the given data dictionary.  This function will only return 
 *              definitions for ODL aggregates. See synGetAggDef for the 
 *              definitions of ODL aggregates.
 *
 * PARAMETERS:  
 *     pDD    -- (synDD) The dictionary containing the desired definition.
 *     pIndex -- (char *) The term for which a definition is required.
 *	   pParentDef -- (synAGGDEF) Definition of the parent aggregate. It's
 *                used to disambiguate identifiers.
 *
 * RETURN VALUE:
 *     synGetParDef -- (synPARDEF) Pointer to the desired definition or NULL.
 *     
 * GLOBAL VARIABLES: None.
 *	
 * NOTES:
 *   1. More than one definition for an ODL identifier may exist in the DD.
 *      If so, the various definitions will be listed under identifiers formed 
 *      by adding ~n (where n is an integer) to the original (ambiguous)
 *      identifier.
 *   2. The ~n discriminators only appear in the DD, never in the actuall ODL.
 *   3. When a DD entry is built for an aggregate, the names of any of its 
 *      components containing ~n discriminators are put into a lookup table 
 *      indexed by the ambiguous part of the name. (E.g., MEDIA_TYPE~1 would be
 *      indexed by MEDIA_TYPE.) 
 *   4. A definition that causes two entries for the same index is considered
 *      ambiguous and cannot be handled. 
 *   5. Only the names of ambiguous componentidentifiers are included in the 
 *      LUT, so a failure to find an entry just means that the identifier is 
 *      not ambiguous.
 *
 * REVISION HISTORY:
 *   10/28/98  Ken Jarva  --  Cloned from synGetAggDef.
 *   11/23/98  Ken Jarva  --  Revised to handle ambiguity.
 *   12/03/98  Ken Jarva  --  Simplified using pParentDef->pAmbiguous.
 *
 ****************************************************************************/
synPARDEF  synGetParDef( synDD pDD, char *pIndex, synAGGDEF pParentDef )
{
	synENTRYDD pEntry;
	synPARDEF pReturn = NULL;
	char *pUnambiguousName; /* Name with the definition 
                                 * discrimiator (~n) */

	/* Make sure the arguments point at something: */
	assert( pDD );
	assert( pIndex );

	/* Determine the unambiguous name if possible */
	if( pParentDef ) {
		/* A definition for the parent aggregate exists, so... */
		pUnambiguousName = sltLookUp( pParentDef->pAmbiguous, pIndex );
		if( !pUnambiguousName ) pUnambiguousName = pIndex;
	} else
		/* No such definition exists, so assume *pIndex is unambiguous: */
		pUnambiguousName = pIndex;

	/* Now get the definition from the data dictionary: */
	pEntry = synDDLookUp( pDD, pUnambiguousName );

	/* If we got one, return the pointer; otherwise return NULL */
	if( pEntry && pEntry->kind == synPARAMETER ){
		pReturn = pEntry->pDef; /* The definition from the DD */
	} else pReturn = NULL;

	return pReturn;

} /* end  synGetParDef */

/*****************************************************************************
 * 
 * NAME: synReportAggResult
 *
 * DESCRIPTION: Reports the result of a syntax check of an ODL aggregate. This
 *              function is silent regarding aggregates that successfully
 *              passed the syntax check.
 *
 * PARAMETERS:  
 *     pF                -- (FILE *) Points at the file to which the data is
 *                                   to be output.
 *     pAggResult        -- (synAGGRESULT) The structure holding the results
 *                                   of the syntax check.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Note the dependence on the definition of synSTATUS enumeration.
 *        The last item in the enumeration should always be synUNDEFINED.
 *
 * REVISION HISTORY:
 *   10/30/98  Ken Jarva  --  Original
 *   11/03/98  Ken Jarva  --  Cleanup and doc header.
 *   12/21/98  Ken Jarva  --  Added code to report on lower level parameters &
 *                            aggregates.
 *
 ****************************************************************************/
void synReportAggResult( FILE *pF, synAGGRESULT pAggResult )
{
	/* note the dependence on the definition of synSTATUS */
	char *status[ synUNDEFINED + 1] = { "PASS", "FAIL", "UNDEFINED" };
	lstLIST pList = NULL;
	synPARRESULT pParResult;
	synAGGRESULT pChildResult;

	assert( pF );
	assert( pAggResult );

	/* Only report problems...*/
	if( (pAggResult->status == synPASS) && (pAggResult->badChildren == 0) ) 
		return;

	/* Report the name and the status: */
	fprintf ( pF, "Syntax Check result for aggragate %s; status is %s\n",
		pAggResult->pAggName, status[ pAggResult->status ]);

	/* Report aggregates that are not of the expected kind: */
	if( pAggResult->wrongKind )
		fprintf( pF, "\tAggreagte is wrong kind.\n" );

	/* Report aggregates that have children that failed their syntax check: */
	if( pAggResult->badChildren )
		fprintf( pF, "\tComponent has problem components.\n" );

	/* Report any missing components: */
	pList = pAggResult->pMissingComponents;
	if( lstNumEl(pList) > 0 ){
		fprintf( pF, "\tThere are the following missing components:\n" );
		lstPrintStringList( pF, pList, "\t\t" );
	}

	/* Report any components that are redundant: */
	pList = pAggResult->pTooManyOccurrences;
	if( lstNumEl(pList) > 0 ){
		fprintf( pF, "\tThere are the following components are redundant:\n" );
		lstPrintStringList( pF, pList, "\t\t" );
	}

	/* Report any unsatisfied selections: */
	pList = pAggResult->pUnsatisfiedSelections;
	if( lstNumEl(pList) > 0 ){
		fprintf( pF, "\tThere are no selections from among the following:\n" );
		lstPrintStringList( pF, pList, "\t\t" );
	}

	/* Report any unknown parameters: */
	pList = pAggResult->pUnknownParameters;
	if( lstNumEl(pList) > 0 ){
		fprintf( pF, "\tThere are unknown parameters specified:\n" );
		lstPrintStringList( pF, pList, "\t\t" );
	}

	/* Report any unknown aggregates: */
	pList = pAggResult->pUnknownAggregates;
	if( lstNumEl(pList) > 0 ){
		fprintf( pF, "\tThere are unknown aggregates specified:\n" );
		lstPrintStringList( pF, pList, "\t\t" );
	}

	/* Report the results of checks on this aggregate's parameters: */
	lstStopIteration( pAggResult->pParameterReports);
	pParResult = lstIterator( pAggResult->pParameterReports );
	while( pParResult ) {
		synReportParResult( pF, pParResult );
		pParResult = lstIterator( pAggResult->pParameterReports );
	}
	lstStopIteration( pAggResult->pParameterReports);

	/* Report the results of checks on this aggregate's children: */
	lstStopIteration( pAggResult->pChildrenReports );
	pChildResult = lstIterator( pAggResult->pChildrenReports );
	while( pChildResult ) {
		synReportAggResult( pF, pChildResult );
		pChildResult = lstIterator( pAggResult->pChildrenReports );
	}
	lstStopIteration( pAggResult->pChildrenReports );

	fprintf( pF, "\n\n" );
	return;

} /* end synReportAggResult */

/*****************************************************************************
 * 
 * NAME: synAggChecker
 *
 * DESCRIPTION: Checks the given aggregate against the given definition and
 *              identifies problems in a report sturcture.
 *
 * PARAMETERS:  
 *     pDef              -- (synAGGDEF ) Points at the definition against  
 *                                       which thegiven aggregate will be 
 *                                       checked.
 *     pAgg              -- (AGGREGATE) Points at he aggreate to be checked.
 *     pResult           -- (synAGGRESULT) Points at the report structure 
 *                                       that's to be used to report the
 *                                       result of the check.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: No definition is given. POST: The result is marked undefined
 *        and no further checks are made.
 *     2. PRE: An aggregate that matches the given definition is provided.
 *        POST: The result is marked synPASS; the lists in the result struct
 *        are all empty and the problem indicators (e.g., pResult->wrongKind )
 *        are all set to 0.
 *     3. PRE: An aggregate that does not match the given definition is 
 *        provided. POST: the result struct is marked synFAIL and an 
 *        indication of the problem is given. The indication may be that one
 *        or more of the indicators is set to 1 or that one or more of the 
 *        lists are not empty.  The checks that are made are as follows:
 *				* The kind of the aggregate (OBJECT or GROUP) must match
 *                the specification in the definition.
 *				* Each element of the ExactlyOne list from the definition
 *                must occur exactly once in the given aggregate. If the
 *                element does not occur, its name goes on the Missing-
 *                Component list; if it occurs more than once it goes on
 *				  TooManyOccurrences.
 *              * Each element of the OneOrMore list must occur at least once
 *                in the aggregate, if not its name is added to Missing-
 *                Components.
 *              * Each element of the ZeroOrOne list may occur at most once;
 *                if it occurs more than once it's name is added to TooMany-
 *                Occurrences.
 *				* Each required select list must have exactly one of its
 *                elements represented one time in the given aggregate.
 *				  << In the  future it will be necessary to recognize lists
 *                    from which the single selected item can appear more 
 *					 than once. >>
 *                If the select list is unrepresented, its contents are
 *                appended to the MissingComponents list; if the list is
 *                represented more than once, its contents are appended to
 *				  the TooManyOccurrences list.
 *              * Each optional select list may have at most one of its
 *                elements represented one time in the given aggregate. 
 *				  << In the future it will be necessary to recognize lists 
 *                   from which the single selected item can appear more 
 *                   than once. >>
 *                if the list is represented more than once, its contents 
 *                are appended to the TooManyOccurrences list.
 *              * Any paramters present in the given aggregate that are not
 *                named in the given definition are listed in the Unknown-
 *                Paramters list.
 *              * Any (child) aggregates present in the given aggregate that
 *                are not named in the given definition are listed in the 
 *				  UnknownAggregates list.
 *
 * NOTES:
 *     1. This function is too long; it needs to be broken up.
  *
 * REVISION HISTORY:
 *   10/30/98  Ken Jarva  --  Original
 *   11/03/98  Ken Jarva  --  Cleanup and doc header.
 *   12/01/98  Ken Jarva  --  Added code to disambiguate the lists.
 *
 ****************************************************************************/
void synAggChecker( synAGGDEF pDef, AGGREGATE pAgg, synAGGRESULT pResult )
{
	lstLIST pChildList, pParameterList, pDefList, pSelectList, pKnownComponents;
	AGGREGATE pChild;
	PARAMETER pParm;
	char *pCompName;

	/* Check the input data: */
	assert( pAgg );
	assert( pResult );

	/* Get the name of the aggregate being checked: */
	pResult->pAggName = pAgg->name;

	/* Make sure it's defined: */
	if( pDef ) {
		pResult->status = synPASS; /* Unless proven otherwise */
	} else{
		pResult->status = synUNDEFINED;
		return;
	}

	/* Initialize the known components list: */
	pKnownComponents = lstInitList();
	
	if( pAgg->kind != pDef->aggKind ) {
		pResult->wrongKind = 1;
		pResult->status = synFAIL;
		return;
	}

	/* Build list of children of the given aggregate: */
	pChildList = lstInitList();
	pChild = pAgg->first_child;
	while( pChild ) {
		char 
			*pName, /* the name as found in the ODL */
			*pUnambigName; /* the unambigous name as derived from the DD
						    * definition */
		pName = pChild->name;
		pUnambigName = sltLookUp( pDef->pAmbiguous, pName );
		if( pUnambigName ) pName = pUnambigName;
		if( pName ) lstAddLast( pChildList, (void *) pName );
		pChild = pChild->right_sibling;
	}
 
	/* Build list of the parameters of the given aggregate: */
	pParameterList = lstInitList();
	pParm = pAgg->first_parameter;
	while( pParm ) {
		char 
			*pName, /* the name as found in the ODL */
			*pUnambigName; /* the unambigous name as derived from the DD
						    * definition */
		pName = pParm->name;
		pUnambigName = sltLookUp( pDef->pAmbiguous, pName );
		if( pUnambigName ) pName = pUnambigName;
		if( pName ) lstAddLast( pParameterList, (void *) pName );
		pParm = pParm->right_sibling;
	}

	/* Check the ExactlyOne list */
	pDefList = pDef->pExactlyOne;
	lstAppendTo( pKnownComponents, pDef->pExactlyOne ); /* for use later */
	lstStopIteration( pDefList );
		pCompName = lstIterator( pDefList );
		while( pCompName ) {
			int count = 0;
			count = lstNumTimesInList( pChildList, pCompName ) +
				lstNumTimesInList( pParameterList, pCompName );
			if( count == 0 ) {
				/* add component's name to the missing components list */
				lstAddLast( pResult->pMissingComponents, strdup(pCompName) );
				pResult->status = synFAIL;
			}
			else if( count > 1) {
				lstAddLast( pResult->pTooManyOccurrences, strdup(pCompName) );
				pResult->status = synFAIL;
			}
			pCompName = lstIterator( pDefList );
		}
	lstStopIteration( pDefList );

	/* Check the OneOrMore list */
	pDefList = pDef->pOneOrMore;
	lstAppendTo( pKnownComponents, pDef->pOneOrMore ); /* for use later */
	lstStopIteration( pDefList );
		pCompName = lstIterator( pDefList );
		while( pCompName ) {
			int count = 0;
			count = lstNumTimesInList( pChildList, pCompName ) +
				lstNumTimesInList( pParameterList, pCompName );
			if( count == 0 ) {
				/* add component's name to the missing components list */
				lstAddLast( pResult->pMissingComponents, strdup(pCompName) );
				pResult->status = synFAIL;
			}
			pCompName = lstIterator( pDefList );
		}
	lstStopIteration( pDefList );

	/* Check the ZeroOrOne list */
	pDefList = pDef->pZeroOrOne;
	lstAppendTo( pKnownComponents, pDef->pZeroOrOne ); /* for use later */
	lstStopIteration( pDefList );
		pCompName = lstIterator( pDefList );
		while( pCompName ) {
			int count = 0;
			count = lstNumTimesInList( pChildList, pCompName ) +
					lstNumTimesInList( pParameterList, pCompName );
			if( count > 1 ) {
				/* add component's name to the TooManyOccurrences list */
				lstAddLast( pResult->pTooManyOccurrences, strdup(pCompName) );
				pResult->status = synFAIL;
			}
			pCompName = lstIterator( pDefList );
		}
	lstStopIteration( pDefList );

	/* Check each of the required selection lists: */
	pDefList = pDef->pReqSelectLists; /* The list of required selection lists */
	lstStopIteration( pDefList );
		pSelectList = lstIterator( pDefList );
		while( pSelectList ) {
			lstLIST pTempList;
			int total = 0; /* the total number of selections from this list */
			/* Check each component in this select list: */
			lstAppendTo( pKnownComponents, pSelectList ); /* for use later */
			pTempList = lstInitList();
			lstStopIteration( pSelectList );
				pCompName = lstIterator( pSelectList ); /* First component in list */
				while( pCompName ) {
					int count = 0; /* how many times this component has been selected */
					count = lstNumTimesInList( pChildList, pCompName ) +
							lstNumTimesInList( pParameterList, pCompName );
					if( count >= 1 ) lstAddLast( pTempList, pCompName );
					total += count; 
					pCompName = lstIterator( pSelectList ); /* Next component */
				}
			lstStopIteration( pSelectList ); /* End iteration on select list */

			if( total > 1 ){
				/* Too many items have been selected from the selction list...*/
				pResult->status = synFAIL;
				lstAppendTo( pResult->pTooManyOccurrences, pTempList );
			} else if( total == 0 ) {
				/* No items have been selected from the selection list...*/
				pResult->status = synFAIL;
				lstAppendTo( pResult->pUnsatisfiedSelections, pSelectList );
			}
			pTempList = lstFreeList( pTempList );

			pSelectList = lstIterator( pDefList ); /* Next select list */
		}
	lstStopIteration( pDefList ); /* End iteration on the list of select lists */


	/* Check each of the optional selection lists: */
	pDefList = pDef->pOptSelectLists; /* The list of required selection lists */
	lstStopIteration( pDefList );
		pSelectList = lstIterator( pDefList );
		while( pSelectList ) {
			lstLIST pTempList;
			int total = 0; /* the total number of selections from this list */
			/* Check each component in this select list: */
			lstAppendTo( pKnownComponents, pSelectList ); /* for use later */
			pTempList = lstInitList();
			lstStopIteration( pSelectList );
				pCompName = lstIterator( pSelectList ); /* First component in list */
				while( pCompName ) {
					int count = 0; /* how many times this component has been selected */
					count = lstNumTimesInList( pChildList, pCompName ) +
							lstNumTimesInList( pParameterList, pCompName );
					if( count >= 1 ) lstAddLast( pTempList, pCompName );
					total += count; 
					pCompName = lstIterator( pSelectList ); /* Next component */
				}
			lstStopIteration( pSelectList ); /* End iteration on select list */

			if( total > 1 ){
				/* Too many items have been selected from the selction list...*/
				pResult->status = synFAIL;
				lstAppendTo( pResult->pTooManyOccurrences, pTempList );
			}
			pTempList = lstFreeList( pTempList );

			pSelectList = lstIterator( pDefList ); /* Next select list */
		}
	lstStopIteration( pDefList ); /* End iteration on the list of select lists */

	/* Finish off the known components list: */
	lstAppendTo( pKnownComponents, pDef->pZeroOrMore ); /* for use later */

	/* Check for unknown parameters: */
	lstStopIteration( pParameterList );
		pCompName = lstIterator( pParameterList );

		while( pCompName ) {
			int count = 0;
			count = lstNumTimesInList( pKnownComponents, pCompName );
			if( count == 0 ) {
				lstAddLast( pResult->pUnknownParameters, pCompName );
				pResult->status = synFAIL;
			}
			pCompName = lstIterator( pParameterList );
		}
	lstStopIteration( pParameterList );


	/* Check for unknown children: */
	lstStopIteration( pChildList );
		pCompName = lstIterator( pChildList );

		while( pCompName ) {
			int count = 0;
			count = lstNumTimesInList( pKnownComponents, pCompName );
			if( count == 0 ) {
				lstAddLast( pResult->pUnknownAggregates, pCompName );
				pResult->status = synFAIL;
			}
			pCompName = lstIterator( pChildList );
		}
	lstStopIteration( pChildList );

	pChildList = lstFreeList( pChildList );
	pParameterList = lstFreeList( pParameterList );
	pKnownComponents = lstFreeList( pKnownComponents );

	return;

} /* end synAggChecker */

/*****************************************************************************
 * 
 * NAME: synReportParResult
 *
 * DESCRIPTION: Prints (via fprint) a brief report to the given file based
 *              on the given syntax check result structure.
 *
 * PARAMETERS:  
 *     pF -- (FILE *) Pointer to the file to which the result is to be 
 *                    written.
 *     pParResult -- (synPARRESULT) The result of a syntax check of a
 *                   parameter.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * REVISION HISTORY:
 *   12/04/98  Ken Jarva  --  Added header to existing function.
 *
 ****************************************************************************/
void synReportParResult( FILE *pF, synPARRESULT pParResult )
{
	char *defStatus = "";
	if( pParResult->status == synPASS ) return;
	else {
		if( pParResult->status == synUNDEFINED ) 
			defStatus = strdup(" is undefined" );
		fprintf( pF, "Bad parameter: %s%s\n", 
			pParResult->pName, defStatus );
		if( pParResult->badType ) 
			fprintf( pF, "\tunexpected data type.\n" );
		if( pParResult->badKind ) 
			fprintf( pF, 
			"\tunexpected kind of value (i.e., scalar, sequence, or set).\n" );
		if( pParResult->outOfRange ) 
			fprintf( pF, "\tvalue is out of range.\n" );
	}
	return;

} /* end synReportParResult */
/*****************************************************************************
 * NAME: synParChecker
 *
 * DESCRIPTION: Checks a given ODL parameter against a given definition.
 *
 * PARAMETERS:  
 *     pDef    -- (synPARDEF) The parameter's definition.
 *     pParm   -- (PARAMETER) The ODL parameter.
 *     pResult -- (synPARRESULT) The strcut used to report the result
 *                of the check of *pParm against *pDef.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Note our dependence on the definition of VALUE_TYPE in odldef.h 
 *        (i.e., see definintion of pTypes[] below)
 *     2. There is no range checking provided.
 *     3. The parameter's kind is checked against that given in the 
 *        definition. An exact match is not required: a scalar is considered
 *        as a special case of sequence or set.
 *
 * REVISION HISTORY:
 *   12/04/98  Ken Jarva  --  Added header to existing code.
 *
 ****************************************************************************/
void synParChecker( synPARDEF pDef, PARAMETER pParm, synPARRESULT pResult )
{
	lstLIST pListOfTypes;
	VALUE_TYPE valType;
	char *pTypes[ TV_DATE_TIME + 1] = { "NULL", "INTEGER", "REAL", "SYMBOL",
		"STRING", "DATE", "TIME", "DATE_TIME" }; /* See odldef.h: VALUE_TYPE */
	VALUE pVal;
	assert( pParm );
	assert( pResult );

	/* Check the input data: */
	if( !pDef ) {
		pResult->status = synUNDEFINED;
		pResult->pName = pParm->name;
		return;
	}

	/* From the definition, get the list of data types the items of this parameter
	 * can take on: */
	pListOfTypes = pDef->pValType;

	/* Get the name of the parameter being checked: */
	pResult->pName = pParm->name;

	/* Assume it's OK until proven otherwise: */
	pResult->status = synPASS;

	/* Check the kind of parameter found against what's expected: */
	if( pParm->value_kind != pDef->valKind ){
		/* Not the expected kind, but treat a scalar as a special case
		 * of sequence or of set */
		if( pParm->value_kind != KV_SCALAR ||
			(pDef->valKind != KV_SEQUENCE && pDef->valKind != KV_SET) ) {
			char msg[256];
			sprintf( msg, "Value-kind mismatch for %s: got %d, expected %d", 
				pParm->name, pParm->value_kind, pDef->valKind );
			synUserWarning( "synParChecker", msg ); 
			pResult->badKind = 1;
			pResult->status = synFAIL;
		}
	}
	
	/* For each value associated with the given parameter, check whether
	 * its type is in the list from the list of allowable types that was 
	 * given in the definition. */
	pVal = pParm->first_value;
	while( pVal ) {
		valType = pVal->item.type;
		if( lstIsStrInList(pListOfTypes, pTypes[valType]) == 0 ){
			/* The item's type is not in the list of allowed types..." */
			pResult->badType = 1;
			pResult->status = synFAIL;
		} 
		pVal = pVal->right_sibling; /* next value node */
	}

	/* Code to check range is TBD */

	return;

 } /* end synParChecker */

/*****************************************************************************
 * 
 * NAME: synDeepAggChecker 
 *
 * DESCRIPTION: Does a deep check of an ODL aggregate; i.e., it first performs
 *		a check of the given aggregate against the given definition, then
 *      it checks each componenet (parameters and child aggregates ) against
 *      their definitions by calling itself recursively.
 *
 * PARAMETERS:  
 *     pResult -- (synAGGRESULT) Pointer to the result struct to be used for
 *                    the check on the given aggregate.
 *     pAgg -- (AGGREGATE) The aggregate to be checked.
 *     pDef -- (synAGGDEF) The definition against which *pAgg is to be checked.
 *     pDD  -- (synDD) The data dictionary. This is needed so that definitions
 *                    can be extracted for lower level components.
 *
 * RETURN VALUE:
 *     pResult -- (synAGGRESULT ) The result struct for the check of *pAgg
 *                    against *pDef. It's also affected by the results of
 *                    the checks done of its *pAgg's children.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This function is recursive.
 *     2. This function reports the results of the checks against *pAgg's
 *        children, but doesn't report on the results of the check done
 *        of *pAgg itself.
 *
 * REVISION HISTORY:
 *   12/04/98  Ken Jarva  --  Added header to existing code.
 *
 ****************************************************************************/
void synDeepAggChecker(synAGGRESULT pResult, AGGREGATE pAgg, 
					   synAGGDEF pDef, synDD pDD )
{
	PARAMETER pParm;
	AGGREGATE pChild;
	synPARDEF pParDef;
	synAGGDEF pAggDef;
	synPARRESULT pParResult;
	synAGGRESULT pAggResult;
	int badComponent = 0;

	/* Check the given aggregate: */
	synAggChecker( pDef, pAgg, pResult );

	/* For each parameter of *pAgg, check the parameter & report result */
	pParm = pAgg->first_parameter;
	while( pParm ) {
		pParDef = synGetParDef( pDD, pParm->name, pDef );
		pParResult = synInitParResult();
		synParChecker( pParDef, pParm, pParResult );
		lstAddLast( pResult->pParameterReports, pParResult );
		if( pParResult->status != synPASS ) badComponent = 1;
		pParm = pParm->right_sibling;
	}

	/* For each child of *pAgg, check the aggregate & report the result */
	pChild = pAgg->first_child;
	while( pChild ) {
		pAggDef = synGetAggDef( pDD, pChild->name , pDef );
		pAggResult = synInitAggResult();
		synDeepAggChecker( pAggResult, pChild, pAggDef, pDD );
		if( pAggResult->status != synPASS  || pAggResult->badChildren ) 
			badComponent = 1;
		lstAddLast( pResult->pChildrenReports, pAggResult );
		pChild = pChild->right_sibling;
	}
	if( badComponent) pResult->badChildren = 1;

	return;

} /* end synDeepAggChecker */

/*****************************************************************************
 * 
 * NAME: synBuildParDef
 *
 * DESCRIPTION: Constructs an internal representation of the definition of a
 *     particular ODL parameter from the ODL representation.
 *
 * PARAMETERS:  
 *     pODLDef -- (AGGREGATE) The definition of a parameter in ODL format.
 *
 * RETURN VALUE:
 *     synBuildParDef -- (synPARDEF) Pointer to the internal representation
 *                        of the defintion given in *pODLDef.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   12/04/98  Ken Jarva  --  Added header to existing code.
 *
 ****************************************************************************/
synPARDEF synBuildParDef( AGGREGATE pODLDef )
{
	PARAMETER pParm;
	synPARDEF pDef;
	struct ODL_Node *pNode;

	pDef = synInitParDef(); /* allocate & initialize a parameter definition
							  * structure */

	/* The Name of the thing being defined must be given... */
	pParm = FindParameter( pODLDef, "NAME" );
	if( pParm ) pDef->pName = pParm->first_value->item.value.string;
	else {
		synFatalError("synBuildParDef",
			"Bad data dictionary. No NAME in par. def." );
	}

	/* The Kind of the thing being defined must be given... */
	pParm = FindParameter( pODLDef, "KIND" );
	if( pParm ) {
		char *pKind;
		pKind = pParm->first_value->item.value.string;
		if( strToUpStrcmp( pKind, "ATTRIBUTE" ) == 0 ) 
			pDef->parKind = KP_ATTRIBUTE;
		else if( strToUpStrcmp( pKind, "POINTER" ) == 0 ) 
			pDef->parKind = KP_POINTER;
		else synFatalError( "synBuildParDef", 
			"definition for unknown parameter kind.");
	} else {
		synFatalError("synBuildAggDef",
			"Bad data dictionary. No KIND in par. def." );
	}

	/* The Value Kind of the thing being defined must be given... */
	pParm = FindParameter( pODLDef, "VALUEKIND" );
	if( pParm ) {
		char *pKind;
		pKind = pParm->first_value->item.value.string;
		if( strToUpStrcmp( pKind, "UNKNOWN" ) == 0 ) 
			pDef->valKind = KV_UNKNOWN;
		else if( strToUpStrcmp( pKind, "SCALAR" ) == 0 ) 
			pDef->valKind = KV_SCALAR;
		else if( strToUpStrcmp( pKind, "SEQUENCE" ) == 0 )
			pDef->valKind = KV_SEQUENCE;
		else if( strToUpStrcmp( pKind, "SET" ) == 0 ) 
			pDef->valKind = KV_SET;
		else synFatalError( "synBuildParDef", "definition for unknown ValueKind.");
	} else {
		synFatalError("synBuildParDef",
			"Bad data dictionary. No ValueKind in parameter def." );
	}

	/* The data type of the thing being defined must be given... */
	/* ***Notice a parameter sequence or set may contain more than one data type, */
	pParm = FindParameter( pODLDef, "TYPE" );
	if( pParm ) {
		pNode = odlMakeNode( PARAMETER_NODE, pParm );
		pDef->pValType = odlGetValueList( pNode );
	} else {
		synFatalError("synBuildParDef",
			"Bad data dictionary. No Type in parameter def." );
	}

	return pDef;

} /* end synBuildParDef */

/*****************************************************************************
 * 
 * NAME: synFreeAggDef
 *
 * DESCRIPTION: Frees an instance of the structure used to store a the 
 *              internal representation of the definition of a particular
 *              ODL aggregate.
 *
 * PARAMETERS:  
 *     pDef  -- (synAGGDEF) Pointer to the aggregate definition structure
 *              that's to be freed.
 *
 * RETURN VALUE: None
 *     
 *
 * NOTES:
 *     1.  It's assumed that pDef really points at a valid structure and
 *     that no other active pointers references any of the lists therein
 *     and that pAggName points at a its own copy of the aggregate name.
 *
 * REVISION HISTORY:
 *   02/17/99  Ken Jarva  --  Original
 *
 ****************************************************************************/

void synFreeAggDef( synAGGDEF pDef )
{
  if( pDef) {

    if( pDef->pAggName ) free( pDef->pAggName );
    pDef->pExactlyOne = lstFreeList( pDef->pExactlyOne );
    pDef->pOneOrMore  = lstFreeList( pDef->pOneOrMore );
    pDef->pOptSelectLists = lstFreeList( pDef->pOptSelectLists );
    pDef->pReqSelectLists = lstFreeList( pDef->pReqSelectLists );
    pDef->pZeroOrMore = lstFreeList( pDef->pZeroOrMore );
    pDef->pZeroOrOne = lstFreeList( pDef->pZeroOrOne );
    pDef->pAmbiguous = lstFreeList( pDef->pAmbiguous );
    
    free( pDef );
  }
  return;
}

/*****************************************************************************
 * 
 * NAME: synInitAggDef
 *
 * DESCRIPTION:  Initialize the structure used to store the internal 
 *               representation of the definition of a particular ODL 
 *               aggregate.
 *
 * PARAMETERS:  None
 *
 * RETURN VALUE:
 *      synInitAggDef -- (synAGGDEF) Pointer to a struct malloc'd on the heap.
 *                       The pointers in the struct have been initialized:
 *                       string pointers to NULL, lstLISTs to empty list
 *                       structs.
 *     
 * GLOBAL VARIABLES: None
 *	
 * REVISION HISTORY:
 *   02/18/99  Ken Jarva  --  Added header.
 *
 ****************************************************************************/
synAGGDEF synInitAggDef( void )
{
	synAGGDEF pDef;
	pDef = malloc( sizeof(*pDef) );

	pDef->pAggName = (char *)NULL;
	pDef->aggKind = KA_OBJECT;
	pDef->ambiguousDef = 0;
	pDef->pExactlyOne = lstInitList();
	pDef->pOneOrMore = lstInitList();
	pDef->pOptSelectLists = lstInitList();
	pDef->pReqSelectLists = lstInitList();
	pDef->pZeroOrMore = lstInitList();
	pDef->pZeroOrOne = lstInitList();
	pDef->pAmbiguous = sltInitLUT();

	return pDef;
}

/*****************************************************************************
 * 
 * NAME: synFreeParDef
 *
 * DESCRIPTION: Frees an instance of a structure that contains the internal
 *              representation of the definition of a particular ODL parameter.
 *
 * PARAMETERS: 
 *     pDef -- (synPARDEF) Pointer to the structre that's to be freed.
 *
 * RETURN VALUE: None
 *
 * REVISION HISTORY:
 *   01/18/99  Ken Jarva  --  Added header.
 *
 ****************************************************************************/
void synFreeParDef( synPARDEF pDef )
{
	if( pDef ){
	    if( pDef->pName ) free( pDef->pName );
	    pDef->pValType = lstFreeList( pDef->pValType );
	    free( pDef );
	}
	return;
}

/*****************************************************************************
 * 
 * NAME: synInitParDef
 *
 * DESCRIPTION: Initializes an instance of a structure that contains the 
 *              internal  representation of the definition of a particular ODL
 *               parameter.
 *
 * RETURN VALUE:  
 *    synInitParDef -- (synPARDEF) Pointer to the structre that's been malloc'd and
 *             initialized.
 *
 * RETURN VALUE: None
 *
 * REVISION HISTORY:
 *   01/18/99  Ken Jarva  --  Added header.
 *
 ****************************************************************************/

synPARDEF synInitParDef( void )
{
	synPARDEF pDef;
	pDef = malloc( sizeof(*pDef) );

	pDef->haveRangeInfo = 0;
	pDef->parKind = KP_ATTRIBUTE;
	pDef->pName = (void *)NULL;
	pDef->pValType = lstInitList();
	pDef->valKind = KV_UNKNOWN;
	
	return pDef;
}

