/* FILE: synODL.h
 *
 * PROGRAM: General use (originally developed for the User Friendly Metadata
 *          (UFM) task).
 *
 * DESCRIPTION: Header file for utility functions used to check ODL
 *		structures against definitions.
 *
 * RCS INFO:
 *
 *	$Id: synODL.h,v 1.2 2000/04/13 00:08:07 ToolsDev Exp $
 *
 *	$Log: synODL.h,v $
 *	Revision 1.2  2000/04/13 00:08:07  ToolsDev
 *	*** empty log message ***
 *
 * Revision 1.2  1999/04/06  16:28:14  ToolsDev
 * Promotion of branch 1.1.0 (cleared wlint warnings) for UFM release.
 *
 * Revision 1.1.0.1  1999/02/22  15:49:01  ToolsDev
 * Changed prototypes to eliminate warnings.
 *
 * Revision 1.1  1999/01/06  21:26:56  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
#ifndef _synODL_header
#define _synODL_header

/* INCLUDED FILES: the following included files are here because of
 *     dependencies that exist within this header file.
*/
#include <stdio.h>

#include "odldef.h"
#include "lstList.h"
#include "sltLUT.h"

/* ---------------------------------------------------------------------*/

/* DATA DECLARATIONS & TYPEDEFS:
 *
*/
struct synAggDef {
	/* The internal representation of the definition of a particular ODL
	 * aggregate (*pAggName). */
	char *  pAggName;
	int ambiguousDef; /* Normally 0; set to 1 iff the definition is
					   * ambiguous; that is if more than one definition
					   * might apply to one of the children. */
	AGGREGATE_KIND 
		    aggKind; /* KA_OBJECT or KA_GROUP */
	lstLIST pExactlyOne;
	lstLIST pZeroOrMore;
    lstLIST pOneOrMore;
    lstLIST pZeroOrOne;
	lstLIST pReqSelectLists;  /* A list of the required selection lists */
	lstLIST pOptSelectLists;  /* A list of the optional selection lists */
	sltLUT  pAmbiguous; /* LUT used to disambiguate terms in 
	                     * aggregate being defined. */
};
typedef struct synAggDef *synAGGDEF;

/*              Parameter Definition Structure                          */
struct synParDef {
	/* The internal representation of the definition of a particular ODL
	 * parameter (*pName). */
	char * pName;
	PARAMETER_KIND parKind; /* KP_ATTRIBUTE or KP_POINTER */
	VALUE_KIND valKind; /* KV_UNKNOWN | KN_SCALAR | KV_SEQUENCE |
						  * KV_SET */
    lstLIST pValType; /* List of the type(s) of data conatined in the
					   * assoicated Value_Data structs. The listed item
					   * may be any of: TV_NULL | TV_INTEGER | TV_REAL |
					   * TV_SYMBOL | TV_STRING | TV_DATE | TV_TIME |
					   * TV_DATE_TIME */
	int haveRangeInfo; /* 0 initially; 1 if and when range information
					    * is added to the struct */
	union RANGE {
		struct ODLInteger  intRange[2];  /* min, max value */
		struct ODLReal     realRange[2]; /* min, max value */
		struct ODLDate     dateRange[2]; /* min, max value */
		       lstLIST     pStringList;  /* list of allowed strings */
	}range;
};
typedef struct synParDef *synPARDEF;

/* The following enum is used to indicate whether a given component
 * matches a particular definition: */
enum synStatus { synPASS, synFAIL, synUNDEFINED };
typedef enum synStatus synSTATUS;

/* The following data structure is used to report the details of
 * syntax check on an ODL aggregate: */
struct synAggSynResult {
	char *pAggName; /* The name of the aggregate to which this 
	                 * report applies */
	synSTATUS status;
	int ambiguous; /* 0 -> no; 1 -> yes */
	int badChildren; /* 0 -> no; 1 -> yes */
	int wrongKind; /* 0 -> no; 1 -> yes */
	lstLIST pMissingComponents;
	lstLIST pTooManyOccurrences;
	lstLIST pUnsatisfiedSelections; /* lstList of lstLists */
	lstLIST pUnknownParameters;
	lstLIST pUnknownAggregates;
	lstLIST pChildrenReports;
	lstLIST pParameterReports;
};
typedef struct synAggSynResult *synAGGRESULT;

/* The following data structure is used to report the details of
 * syntax check on an ODL parameter: */
struct synParSynResult {
	char *pName; /* The name of the parameter */
	synSTATUS status;
	int ambiguous; /* 0 -> no; 1 -> yes */
	int badType; /* 0 -> no; 1 -> yes */
	int badKind; /* 0 -> no; 1 -> yes */
	int outOfRange; /* 0 -> no; 1 -> yes */
};
typedef struct synParSynResult *synPARRESULT;

/* The following data structures  comprise the data dictionary: */

struct synDataDictionary{
	/* A DD has 27 lookup tables, one for each letter of the
	 * alphabet and one extra for anything else. The first letter of the
	 * term being defined determines which of the 27 LUTs is used.*/
	sltLUT section[27];
};

typedef struct synDataDictionary *synDD; /* pointer to a data dictionary
								   * structure. */

enum synCompKind { synUnknownCompKind, synAGGREGATE, synPARAMETER };
typedef enum synCompKind synCOMPKIND;

/* The following structure represents an entry in a data dictionary. */
struct synEntryDD {
	synCOMPKIND kind;  /* synAGGREGATE | synPARAMETER */
	void *pDef;         /* synAGGDEF or synPARDEF, depending on kind */
};
typedef struct synEntryDD *synENTRYDD;

/* ---------------------------------------------------------------------*/

/* EXTERN VARIABLES: the following variables may be used in other files.
 *    They are defined in synODL.c.
*/
	/* NONE */


/* ---------------------------------------------------------------------*/

/* EXTERN PROTOTYPES: the following prototypes are for functions defined
 *     in synODL.c but which are intended for use 
 *     elsewhere.
*/
void synInitialization( int debugFlag, int verboseFlag, int noExit,
	     FILE *pDebugFile, FILE *pUserMessages );
synDD synInitDataDict( void );
int   synAddToDD( synDD pDD, char *pName, synENTRYDD pEntry );
synENTRYDD synDDLookUp( synDD pDD, char *pIndex );
int synBuildDataDict( synDD pDD, AGGREGATE pDDTree );
void synFreeParDef( synPARDEF pDef );
void synFreeAggDef( synAGGDEF pDef );
synAGGDEF synGetAggDef( synDD pIFDD, char *pIndex, synAGGDEF pParentDef );
synPARDEF synGetParDef( synDD pDD, char *pIndex, synAGGDEF pParentDef );
void synReportAggResult( FILE *pF, synAGGRESULT pAggResult );
void synPrintAggDef( FILE *pF, synAGGDEF pDef );
void synDeepAggChecker( synAGGRESULT pResult, AGGREGATE pAgg, 
					   synAGGDEF pDef, synDD pDD );
void synAggChecker( synAGGDEF pDef, AGGREGATE pAgg,synAGGRESULT pResult );
void synParChecker( synPARDEF pDef, PARAMETER pPar,synPARRESULT pResult );
void synReportParResult( FILE *pF, synPARRESULT pAggResult );
synAGGRESULT synInitAggResult( void );
synPARRESULT synInitParResult( void );
/* ---------------------------------------------------------------------*/

#endif 
