/* FILE: ufm.c
 *
 * PROGRAM: User Friendly Metadata (UFM)
 *
 * DESCRIPTION: This file contains the main function, and some high-level
 *     supporting functions, for the UFM tool.
 *
 * RCS INFO:
 *
 *      $Id: ufm.c,v 1.3 2000/04/06 23:38:02 ToolsDev Exp $
 *
 *      $Log: ufm.c,v $
 *      Revision 1.3  2000/04/06 23:38:02  ToolsDev
 *      Made changes to display the ODL file header informations:
 *      created new case 'h' to invok the header option in the command
 *      line.
 *      created new variable for the header in ufmControls.
 *      wrote comments to reflect the changes.
 *      (Amer Rahabi)
 *
 * Revision 1.2  1999/04/06  15:51:38  ToolsDev
 * Promotion of branch 1.1.0 (cleared wlint warnings) for release.
 *
 * Revision 1.1.0.2  1999/03/05  18:01:57  ToolsDev
 * *** empty log message ***
 *
 * Revision 1.1.0.1  1999/03/05  16:48:08  ToolsDev
 * Cleared several points that generted lint warnings.
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
/* INCLUDED FILES:
*/
#include "ufmMain.h"
#include "macMacro.h"
#include "sltLUT.h"
#include "odldef.h"
#include "odlinter.h"
#include "cmpIFCompiler.h"
#include "ReportWriter.h"
#include "synODL.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/* ---------------------------------------------------------------------*/

/* GLOBAL VARIABLES: the following variables may be used in other files.
*/
ufmControls ufmCONTROLS; /* UFM controls */
/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
/* Exit status to return to the operating system. */
const int ExitSuccess =  0;
const int ExitFailure = -1;

/* Return status to return to a calling function. */
const int Success = 1;
const int Failure = 0;
/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static int CmdLineProc (int argc, char **argv, ufmControls *ufmCONTROLS);
static void CreateDefInstr (AGGREGATE  rootIF);
int getopt (int argc, char *const argv[], const char *opts);
static void ufmUserWarning (char *functionName, char *msg);
static void ufmFatalError (char *functionName, char *msg);
void getColumn(FILE *insfile, char columns[20][256], int *n);
void writeColumn(FILE *odlfile, FILE *tmpfile, char columns[20][256], int n, char *TARGETG, char *TARGETO);
void LtoU (char *inLine, char *outLine);
/* ---------------------------------------------------------------------*/

/*****************************************************************************
 * 
 * NAME: main
 *
 * DESCRIPTION: This is the main function of the UFM tool.
 *
 * PARAMETERS:
 *     argc (int)              The number of command line arguments.
 *     argv (char **)          An array of character pointers.  Each pointer
 *                             points to one of the command line arguments.
 *
 * RETURN VALUE:
 *     Exit status = ExitSuccess --> The UFM tool completed successfully,
 *                                   although some non-severe errors, or
 *                                   warnings, may have occurred.
 *                 = ExitFailure --> The UFM tool encountered a fatal error
 *                                   and could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS (ufmControls)  A data structure containing various controls
 *                                of the UFM tool.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     10/05/1998  J.D. Stroup  Original
 *     03/05/1999  K.V. Jarva   Cleared several lint warnings.
 *
 ****************************************************************************/

int main(int argc, char **argv)
{
  AGGREGATE  rootIF, rootODL;
  struct REP report;

  int  i, j, k, lineC=256, columnArray[20];
  int  w, x, y, z, pathL, groupL;
  char *insfilename, *odlfilename;
  char aTempLine[256];
  char *tmpfilename="tmpfile", *tstfilename="tstfile";
  FILE *insfile, *odlfile, *tmpfile, *tstfile;
  char oLine[256], aLine[256], addLine[256], groupString[256], tempLine[256];
  char *astring, *bstring, *cstring, *target;
  char tempColumn[100], endGroup[100], endObject[100];
  char newPath[256], path[256];
  char *group, *firstGroup, startFirstGroup[256], endFirstGroup[256], groups[10][100];
  char startFirstObject[256], endFirstObject[256];
  char *newLine, pathAndLine[256];
  int  tn=0, cn[10];
  char targets[10][100];
  char columns[20][20][256];
  char TargetG[100], TargetO[100];


  /* Check for arguments */
  if (argc ==1)
  {
    fprintf (stderr, "\nUsage: ufm [-d|-v|-h|-c <comment>|-f <ODL_file>|"
                     "-g <debug_file>|\n            -i <instructions_file>|"
                     "-o <HTML_file>|-t <title>]\n");
    return (1);
  }

  /* Process the command line options, and initialize the UFM controls. */

  if (CmdLineProc (argc, argv, &ufmCONTROLS) != Success) exit (ExitFailure);

  //Get file names
  for (i=0; i<argc; i++)
  {
    if (strcmp(argv[i], "-i")==0)
    {
      insfilename=argv[i+1];
    }
    if (strcmp(argv[i], "-f")==0)
    {
      odlfilename=argv[i+1];
      argv[i+1]=tmpfilename;
    }
  }
 
  //Open all files
  insfile=fopen(insfilename, "r");
  //tstfile=fopen(tstfilename, "w");
  odlfile=fopen(odlfilename, "r");
  tmpfile=fopen(tmpfilename, "w");

  //Read insfile to get all tables column name.
  while (!feof(insfile))
  {
    if (fgets(oLine, lineC, insfile)==NULL) goto insEnd;

    //To upcase.
    LtoU(oLine, aLine);
   
    //Get any target.
    if (strstr(aLine, "TARGET = ")!=NULL)
    {
      strcpy(aTempLine, aLine);
      target=strtok(aTempLine, "\"");
      target=strtok(NULL, "\"");
    }
   
    //Get table and save latest target.
    if (strstr(aLine, "OBJECT = TABLESPECBYCOLUMN")!=NULL)
    {
      strcpy(targets[tn], target);
      getColumn(insfile, columns[tn], &cn[tn]); 
      tn++;
    }
  }
  insEnd:;

  //Read odlfile and write to tmpfile.
  //If tables column name is null add blank "".
  while (!feof(odlfile))
  {
    if (fgets(oLine, lineC, odlfile)==NULL) goto odlEnd;
    LtoU(oLine, aLine);
    //Write original line. 
    //fputs(aLine, tmpfile);
    fputs(oLine, tmpfile);

    //Get no space line.
    strcpy(tempLine, "");
    astring=strtok(aLine, "\n");
    bstring=strtok(astring, "	");
    astring=strtok(bstring, " ");
    while(astring)
    {
      strcat(tempLine, astring);
      astring=strtok(NULL, " ");
    }
    for (i=0; i<tn; i++)
    {
      strcpy(TargetG, "GROUP=");
      strcpy(TargetO, "OBJECT=");
      strcat(TargetG, targets[i]);
      strcat(TargetO, targets[i]);
      if (strcmp(tempLine, TargetG)==0 || strcmp(tempLine, TargetO)==0)
      {
        writeColumn(odlfile, tmpfile, columns[i], cn[i], TargetG, TargetO);
        break;
      }
    }
  }
  odlEnd:;

  fclose(insfile);
  //fclose(tstfile);
  fclose(odlfile);
  fclose(tmpfile);


    /* Initialize the utility modules: */

  	synInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  	sltInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  	lstInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  /* Create the root node of the instructions file tree. */

  if ((rootIF = NewAggregate (NULL, KA_GROUP, "IF_ROOT", NULL)) == NULL)
    ufmFatalError ("main",
                   "Creation of the instructions file's root node failed");

  /* Read the instructions file, and create its tree under its root node. */

  if (ufmCONTROLS.pInstrFile != NULL)
  {
	  if (ReadLabel (ufmCONTROLS.pInstrFile, rootIF) != Success)
		  ufmFatalError ("main",
                  "ReadLabel failed to create the instructions file's tree.");
  }
  else
		ufmUserWarning ("main",
            "No instructions file.  Default instructions only will be used.");

  /* Add default instructions to ODL instructions tree. */

  CreateDefInstr (rootIF);

  /* Run the first pass compile step. */

	if (cmpPass1 (rootIF) != Success)
		ufmFatalError ("main", "Compiler pass 1 failure");

  /* Create the root node of the input ODL tree. */

	if ((rootODL = NewAggregate (NULL, KA_GROUP, "ROOT", NULL)) == NULL)
    ufmFatalError ("main", "Creation of the input ODL root node failed");

  /* Read the input ODL file, and create its tree under its root node. */

  if (ReadLabel (ufmCONTROLS.pODL_In, rootODL) != Success)
		ufmFatalError ("main",
                   "ReadLabel failed to create the ODL file's tree.");

  /* Run the second pass compile step. */

	if (cmpPass2 (&report, rootODL) != Success)
		ufmFatalError( "main", "Compiler pass 2 failure");

  /* Generate the output, HTML report only if no compiler errors
     encountered. */

  if (! ufmCONTROLS.CompilerFailure) {
    if (ReportWriter (&report) != Success)
		  ufmFatalError ("main", "ReportWriter failure");
    {
      /* On PC, comment next line. */
      system("rm tmpfile");
      exit (ExitSuccess);
    }
  }
  exit (ExitFailure);
  return -1;
}

/*****************************************************************************
 * 
 * NAME: CmdLineProc
 *
 * DESCRIPTION: This function processes the command line options.  It verifies
 *     that the options are valid for the UFM tool.  It also initializes the
 *     UFM control structure to default values, which may be overridden via
 *     command line options.
 *
 * PARAMETERS:
 *     argc (int)              The number of command line arguments.
 *     argv (char **)          An array of character pointers.  Each pointer
 *                             points to one of the command line arguments.
 *     controls (ufmControls)  The UFM's control structure.  It is initialized
 *                             by this function.
 *
 * RETURN VALUE:
 *     CmdLineProc (int)  Success --> This function completed successfully.
 *                        Failure --> This function encountered an error and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Insufficient memory on the heap, and a title or comment option
 *              given on the command line.
 *        POST: Error message generated stating that memory for the title or
 *              comment string could not be allocated.  The function returns a
 *              status of "Failed". 
 *     2. PRE:  An invalid option given on the command line.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *     3. PRE:  An option's argument given that begins with a dash--illegal
 *              syntax.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *     4. PRE:  An option with an argument that is a file name.  The file does
 *              not exist.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *
 * NOTES:
 *     The standard rules for specifying Unix command line options apply here.
 *
 *     Most valid options to the UFM tool have arguments.  The name of an
 *     option's argument can not begin with a dash ("-").  If it does, then
 *     this function considers this an error: a message is generated to
 *     STDERR, and this function returns with a "failed" status.
 *
 *     If the same option is given multiple times on the command line, the
 *     last instance of the option is used.
 *
 * REVISION HISTORY:
 *     10/05/1998  J.D. Stroup  Original
 *
 ****************************************************************************/

static int CmdLineProc (int argc, char **argv, ufmControls *controls)
{
  extern char *optarg;
  extern int  optind;
  extern int  opterr;

  const char *DefaultTitle     = "UFM Output";
  const char *ValidOptions     = "dvhc:f:g:i:o:t:";
  const char Dash              = '-';
  const char *DefaultInstrFile = "ufmInstructions.odl";

  enum {None, InvalidOption, InvalidOptionArg, OpenFailureODL, OpenFailureIF,
        OpenFailureHTML, OpenFailureDebug, MallocFailureComment,
        MallocFailureTitle} error;

  int opt;

  opterr = 0;   /* disable error messages in getopt */
  error = None;

/* Initialize the fields of the UFM's controls structure to their default
   values. */

  controls->pODL_In    = stdin;
  controls->pInstrFile = NULL;
  controls->pInstrFile = fopen (DefaultInstrFile, "r");
  controls->pHTML_Out  = stdout;
  controls->pDebugFile = controls->pCompilerErrors = stderr;
  controls->verbose    = controls->debug = controls->CompilerFailure = 0;
  controls->pTitle     = strdup (DefaultTitle);
  controls->pComment   = NULL;
  controls->DHeader    = 0 ;   
				 

  if (controls->pTitle == NULL) error = MallocFailureTitle;

/* Retrieve and process each option given on the command line.  This continues
   until all options are exhauseted or an error is encountered. */

  while ((error == None) &&
         ((opt = getopt (argc, argv, ValidOptions)) != EOF))
  {
    switch ((char) opt)
    {
      /* Comment option (c).  Its argument is a string of text. */

      case 'c':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
        {
          controls->pComment = strdup (optarg);

          if (controls->pComment == NULL) error = MallocFailureComment;
        }
        break;

      /* Debug option (d).  It has no arguments. */

      case 'd':
        controls->debug = 1;
        break;

      /* Input ODL file option (f).  Its argument is the name of a file
         containing the input ODL. */

      case 'f':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pODL_In = fopen (optarg, "r")) == NULL)
            error = OpenFailureODL;
        break;

      /* Error message file option (g).  Its argument is the name of a file
         to which any error or debug messages are written. */

      case 'g':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pCompilerErrors = fopen (optarg, "w")) == NULL)
            error = OpenFailureDebug;
          else
            controls->pDebugFile = controls->pCompilerErrors;
        break;

	/*  Header Information option (h), copied from the orginal ODL file, It has no argument. */

      case 'h':
	 
	      controls->DHeader    = 1;
          
      
        break;
	

      /* Instruction file option (i).  Its argument is the name of an
         instructions file. */

      case 'i':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pInstrFile = fopen (optarg, "r")) == NULL)
            error = OpenFailureIF;
        break;

      /* Output file option (o).  It argument is the name of a file to which
         the output HTML is written. */

      case 'o':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pHTML_Out = fopen (optarg, "w")) == NULL)
            error = OpenFailureHTML;
        break;

      /* Title option (t).  Its argument is a string of text. */

      case 't':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
        {
          controls->pTitle = strdup (optarg);

          if (controls->pTitle == NULL) error = MallocFailureTitle;
        }
        break;

      /* Verbose option (v).  It has no arguments. */

      case 'v':
        controls->verbose = 1;
        break;

      /* Invalid option or missing mandatory argument of an option. */

      case '?':
        error = InvalidOption;
        break;
    }
  }

  /* If no error were encountered processing the command line options, then
     this function has completed successfully. */

  if (error == None)
    return (Success);

  /* Otherwise, an error has occurred when processing the command line
     options.  Output the appropriate error message to standard error, and
     display the correct usage for the UFM tool. */

  else
  {
    switch (error)
    {
      case InvalidOptionArg:
        fprintf (stderr, "Option argument's name can not begin with a "
                         "dash ('-')\n");
        break;
      case MallocFailureComment:
        fprintf (stderr, "Malloc failure: cannot allocate memory "
                         "for comment string\n");
        break;
      case MallocFailureTitle:
        fprintf (stderr, "Malloc failure: cannot allocate memory "
                         "for title string\n");
        break;
      case OpenFailureODL:
        fprintf (stderr, "Failed to open the given ODL file\n");
        break;
      case OpenFailureIF:
        fprintf (stderr, "Failed to open the given instructions file\n");
        break;
      case OpenFailureHTML:
        fprintf (stderr, "Failed to open the given HTML output file\n");
        break;
      case OpenFailureDebug:
        fprintf (stderr, "Failed to open the given debug/error file\n");
        break;
      case InvalidOption:
        fprintf (stderr, "Invalid option or missing option's argument\n");
        break;
    }
    fprintf (stderr, "\nUsage: ufm [-d|-v|-h|-c <comment>|-f <ODL_file>|"
                     "-g <debug_file>|\n            -i <instructions_file>|"
                     "-o <HTML_file>|-t <title>]\n");
    return (Failure);
  }
} /* end CmdLineProc */

/*****************************************************************************
 * 
 * NAME: CreateDefInstr
 *
 * DESCRIPTION: This function creates the default instructions and hangs them
 *     onto the end (or the right-most siblings) of the instruction file's ODL
 *     tree.
 *
 * PARAMETERS:
 *     rootIF (AGGREGATE)  Pointer to the root node of the instruction file's
 *                         ODL tree.
 *
 * RETURN VALUE:
 *     None.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     There are three default instructions:
 *       1) an instruction for the root node;
 *       2) an instruction for an AGGREGATE node; and
 *       3) an instruction for a PARAMETER node.
 *
 *      The root node instruction is not absolutely necessary since the root
 *      node is an AGGREGATE node, and would be covered by the general
 *      AGGREGATE node instruction.  But a special instruction is used for
 *      root that is not used of other AGGREGATE nodes: the
 *      "ReplaceWithSubstructure" instruction component.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     12/16/1998  Ken Jarva -- set value_kind.
 *     12/17/1998  J.D. Stroup -- set value_count.
 *
 ****************************************************************************/

static void CreateDefInstr (AGGREGATE  rootIF)
{
  AGGREGATE  p_agg, p_comp_agg;
  PARAMETER  p_parm;
  VALUE_DATA *parm_value = (VALUE_DATA *) malloc (sizeof (struct Value_Data));

  /* The following parameter value information remains constant, so it's set
     here once. */

  parm_value->type         = TV_STRING;
  parm_value->valid        = 1;
  parm_value->format       = 0;
  parm_value->precision    = 0;

  /**********************************************************/
  /* Create the "ufmROOT" AggregateTranslation instruction. */
  /**********************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "AGGREGATETRANSLATION", NULL);

  /* target name component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGET");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmROOT");
  parm_value->value.string = strdup ("ufmROOT");

  NewValue (p_parm, parm_value);

  /* list specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "LISTSPEC", NULL);

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "ELEMENTS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CHILDREN)");
  parm_value->value.string = strdup ("$(_CHILDREN)");

  NewValue (p_parm, parm_value);

  /* "replace with substructure" component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "REPLACEWITHSUBSTRUCTURE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("YES");
  parm_value->value.string = strdup ("YES");

  NewValue (p_parm, parm_value);

  /***************************************************************/
  /* Create the "ufmAGGREGATE" AggregateTranslation instruction. */
  /***************************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "AGGREGATETRANSLATION", NULL);

  /* target class component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGETCLASS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmAGGREGATE");
  parm_value->value.string = strdup ("ufmAGGREGATE");

  NewValue (p_parm, parm_value);

  /* list specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "LISTSPEC", NULL);

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "ELEMENTS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CHILDREN)");
  parm_value->value.string = strdup ("$(_CHILDREN)");

  NewValue (p_parm, parm_value);

  /***************************************************************/
  /* Create the "ufmPARAMETER" AggregateTranslation instruction. */
  /***************************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "PARAMETERTRANSLATION", NULL);

  /* target class component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGETCLASS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmPARAMETER");
  parm_value->value.string = strdup ("ufmPARAMETER");

  NewValue (p_parm, parm_value);

  /* "replace with substructure" component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "REPLACEWITHSUBSTRUCTURE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("YES");
  parm_value->value.string = strdup ("YES");

  NewValue (p_parm, parm_value);

  /* assignment specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "ASSIGNMENTSPEC", NULL);

  /* left side */

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "LEFTSIDE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CURRCOMP)");
  parm_value->value.string = strdup ("$(_CURRCOMP)");

  NewValue (p_parm, parm_value);

  /* right side */

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "RIGHTSIDE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_VALUE)");
  parm_value->value.string = strdup ("$(_VALUE)");

  NewValue (p_parm, parm_value);

  return;
} /* end CreateDefInstr */

/*****************************************************************************
 * 
 * NAME: ufmFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void ufmFatalError( char *functionName, char *msg )
{
  fprintf( ufmCONTROLS.pDebugFile, "%s: FATAL - %s\n", functionName, msg );
	exit(ExitFailure);

} /* end ufmFatalError */

/*****************************************************************************
 * 
 * NAME: ufmUserWarning
 *
 * DESCRIPTION: Reports warnings if the verbose flag in the UFM control
 *     structure is set (= 1).
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.verbose  -- (int)  Flag indicating whether the program is
 *                                    in verbose mode (= 1) or not (= 0).
 *     ufmCONTROLS.pCompilerErrors
 *                          -- (FILE) The debug file to which error messages
 *                                    are to be written.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void ufmUserWarning( char *functionName, char *msg )
{
	if( !ufmCONTROLS.verbose ) return;

  fprintf( ufmCONTROLS.pCompilerErrors, "%s: WARNING - %s\n", functionName,
                                                              msg );
	return;

} /* end ufmUserWarning */

void getColumn(FILE *insfile, char columns[20][256], int *n)
{
  int i, j, k, lineC=256;
  char *astring, *bstring, *cstring, oLine[256], aLine[256], tempLine[256]; 

  while(!feof(insfile))
  {
    fgets(oLine, lineC, insfile);
   
    //To upcase.
    LtoU(oLine, aLine);

    //Get columns
    strcpy(tempLine, aLine);
    //Get rid of any tab before COLUMNS.
    astring=strtok(tempLine, "	");
    bstring=strtok(astring, " ");
    if(strcmp(bstring, "COLUMNS")==0)
    {
      //For Columns line. Get first column.
      *n=1;
      cstring=strtok(NULL, "\"");
      cstring=strtok(NULL, "\"");
      strcpy(columns[0], cstring);
      //For other lines. 
      for (i=1; i<20; i++)
      {
        fgets(oLine, lineC, insfile);
        LtoU(oLine, aLine);
        if (strstr(aLine, "END_OBJECT = TABLESPECBYCOLUMN")!=NULL) goto TableEnd;
        *n=*n+1;
        cstring=strtok(aLine, "\"");
        cstring=strtok(NULL, "\"");
        strcpy(columns[i], cstring);
      }
    }
  }
  TableEnd:;
}
      
void writeColumn(FILE *odlfile, FILE *tmpfile, char columns[20][256], int n, char *TARGETG, char *TARGETO)
{
  int  i, j, k, l, a, blankN, lineC=256, columnArray[20];
  int  w, x, y, z, pN, pathL, groupL;
  char oLine[256], aLine[256], addLine[256], groupString[256], tempLine[256], aTempLine[256];
  char *astring, *bstring, *cstring, *blankS;
  char tempColumn[100], endGroup[100], endObject[100];
  char newPath[100], path[100], *seGroup;
  char *group, *firstGroup, startFirstGroup[256], endFirstGroup[256], groups[10][100];
  char startFirstObject[256], endFirstObject[256];
  char *newLine, pathAndLine[256];

  strcpy(endGroup, "END_");
  strcat(endGroup, TARGETG);
  strcpy(endObject, "END_");
  strcat(endObject, TARGETO);

  strcpy(groupString, columns[0]);
  firstGroup=strtok(groupString, "/");
  //printf("%s\n", firstGroup);
  sprintf(startFirstGroup, "GROUP=%s", firstGroup);
  sprintf(endFirstGroup, "END_GROUP=%s", firstGroup);
  sprintf(startFirstObject, "OBJECT=%s", firstGroup);
  sprintf(endFirstObject, "END_OBJECT=%s", firstGroup);

  while(!feof(odlfile))
  {
    //Read and write a line with upcase. 
    fgets(oLine, lineC, odlfile);
    LtoU(oLine, aLine);

    //Write original line. 
    //fputs(aLine, tmpfile);
    fputs(oLine, tmpfile);

    //Make no space line
    strcpy(tempLine, "");
    astring=strtok(aLine, "\n");
    bstring=strtok(astring, "	");
    astring=strtok(bstring, " ");
    while(astring)
    {
      strcat(tempLine, astring);
      astring=strtok(NULL, " ");
    }

    if(strcmp(tempLine, endGroup)==0 ||
       strcmp(tempLine, endObject)==0)
       break; 

    //First group.
    if (strcmp(tempLine, startFirstGroup)==0 ||
        strcmp(tempLine, startFirstObject)==0)
    {
      for (i=0; i<n; i++)
      {  
        columnArray[i]=0;
      }
      blankN=0;
      strcpy(path, firstGroup);
      strcat(path, "/");
      
      //Read this group. 
      while(!feof(odlfile))
      {
        fgets(oLine, lineC, odlfile);
        LtoU(oLine, aLine);
        
        //Make no space line
        strcpy(aTempLine, aLine);

        strcpy(tempLine, "");
        astring=strtok(aTempLine, "\n");
        bstring=strtok(astring, "	");
        astring=strtok(bstring, " ");
        while(astring)
        {
          strcat(tempLine, astring);
          astring=strtok(NULL, " ");
        }
        //If not endFirstGroup. 
        if (strcmp(tempLine, endFirstGroup)!=0 &&
            strcmp(tempLine, endFirstObject)!=0) 
        {
          //Start Group.
          strcpy(tempLine, aLine);
          astring=strtok(tempLine, "	");
          seGroup=strtok(astring, " ");

          if (strcmp(seGroup, "GROUP")==0 ||  
              strcmp(seGroup, "OBJECT")==0 ) 
          {
            strcpy(tempLine, aLine);
            astring=strtok(tempLine, "\n");
            group=strtok(astring, " ");
            group=strtok(NULL, " ");
            group=strtok(NULL, " ");

            strcat(path, group);
            strcat(path, "/");
          }
          //End-group.
          if (strcmp(seGroup, "END_GROUP")==0 ||  
              strcmp(seGroup, "END_OBJECT")==0 ) 
          {
            strcpy(tempLine, aLine);
            
            astring=strtok(tempLine, "\n");
            group=strtok(astring, " ");
            group=strtok(NULL, " ");
            group=strtok(NULL, " ");

            pathL=strlen(path);
            groupL=strlen(group);
            strcpy(newPath, "");
            pN=pathL-groupL-1;
            strncpy(newPath, path, pN);
            newPath[pN]='\0';
            strcpy(path, newPath);
          }

          //If not end, write, otherwise, add blank, then write end.
          //Write original line. 
          //fputs(aLine, tmpfile);
          fputs(oLine, tmpfile);

          //Check which column is missing.
          strcpy(tempLine, aLine);
          astring=strtok(tempLine, " ");
          newLine=strtok(astring, "	");
          //printf("%s\n", newLine);
          strcpy(pathAndLine, path);
          strcat(pathAndLine, newLine);
          //printf("%s\n", pathAndLine);
          //Get blank number. 
          if (strcmp("GRINGPOINT/GRINGPOINTSEQUENCENO/NUM_VAL", pathAndLine)==0 ||
              strcmp("GRINGPOINT/GRINGPOINTLATITUDE/NUM_VAL", pathAndLine)==0 ||
              strcmp("GRINGPOINT/GRINGPOINTLONGITUDE/NUM_VAL", pathAndLine)==0)
          {
            strcpy(tempLine, aLine);
            /*
            astring=strtok(tempLine, " ");
            while(astring)
            { 
              strcpy(blankS, astring);
              astring=strtok(NULL, " ");
            }
            */
            blankS=strtok(tempLine, " ");
            blankS=strtok(NULL, " ");
            blankS=strtok(NULL, " ");
            blankN=atoi(blankS);
            //printf("%i\n", blankN);
          }
          for (j=0; j<n; j++)
          {  
            if(strcmp(pathAndLine, columns[j])==0)
            {
              columnArray[j]=1;
              //printf("%s\n", columns[j]);
              break;
            }
          }
        }
        else
        {
          for (j=0; j<n; j++)
          {
            if (columnArray[j]==0)
            {
              //Parse and write group.

              strcpy(tempColumn, columns[j]);
              astring=strtok(tempColumn, "/");
              //Find how many group.
              z=0;
              while(astring)
              {
                z++;
                astring=strtok(NULL, "/");
              }
              //Print group and value.
              strcpy(tempColumn, columns[j]);
              astring=strtok(tempColumn, "/");
              x=0;
              for (w=0; w<z; w++)
              {
                /*
                //Write start groups.
                if (w!=0)
                {
                  strcpy(groups[x], astring);
                  x++;
                  fprintf(tmpfile, "GROUP = %s\n", astring);
                }
                */
                
                //Write value.
                if (w==(z-1))
                {
                  if (blankN==0)
                  {
                    fprintf(tmpfile, "%s=\"\"\n", astring);
                  }
                  else
                  {
                    if (strcmp(astring, "VALUE")==0)
                    {
                      fprintf(tmpfile, "%s=(", astring);
                      for (a=1; a<blankN; a++)
                      {
                        fprintf(tmpfile, "\"\",", astring);
                      } 
                      fprintf(tmpfile, "\"\"", astring);
                      fprintf(tmpfile, ")\n", astring);
                    }
                    else
                    {
                      fprintf(tmpfile, "%s=\"\"\n", astring);
                    }
                  }
                  //printf("%s\n", astring);
                  break;
                }
                //Write start groups. Don't need first group.
                if (w!=0)
                {
                  strcpy(groups[x], astring);
                  x++;
                  fprintf(tmpfile, "GROUP = %s\n", astring);
                }
                //Get another string.
                astring=strtok(NULL, "/");
              }
              //Print end group from different directions. Equal start group. 
              for (y=(x-1); y>=0; y--)
              {
                fprintf(tmpfile, "END_GROUP = %s\n", groups[y]);
              }
            }
          }
          //Write end of first group.
          //Write original line. 
          //fputs(aLine, tmpfile);
          fputs(oLine, tmpfile);
          goto writeCEnd;
        }
      }
    }
    writeCEnd:;
  }   
}

void LtoU(char *inLine, char *outLine)
{
  int i;
  char astring, bstring;
 
  for (i=0; i<strlen(inLine); i++)
  {
    astring=inLine[i];
    bstring=toupper(astring);
    outLine[i]=bstring;
  }
  outLine[strlen(inLine)]='\0';
}
